from configparser import ConfigParser
from importlib import resources
from os.path import basename, splitext

from pyaerocom import const
from pyaerocom.exceptions import FileConventionError
from pyaerocom.tstype import TsType


class FileConventionRead:
    """Class that represents a file naming convention for reading Aerocom files

    Attributes
    ----------
    name : str
        name of this convention (e.g. "aerocom3")
    file_sep : str
        filename delimiter for accessing different variables
    year_pos : int
        position of year information in filename after splitting using
        delimiter :attr:`file_sep`
    var_pos : int
        position of variable information in filename after splitting using
        delimiter :attr:`file_sep`
    ts_pos : int
        position of information of temporal resolution in filename after
        splitting using delimiter :attr:`file_sep`
    vert_pos : int
        position of information about vertical resolution of data
    data_id_pos : int
        position of data ID
    """

    _io_opts = const
    AEROCOM3_VERT_INFO = {
        "2d": ["surface", "column", "modellevel"],
        "3d": ["modellevelatstations"],
    }

    def __init__(
        self,
        name="aerocom3",
        file_sep="_",
        year_pos=None,
        var_pos=None,
        ts_pos=None,
        vert_pos=None,
        data_id_pos=None,
        from_file=None,
    ):

        self.name = name
        self.file_sep = file_sep

        self.year_pos = year_pos
        self.var_pos = var_pos
        self.ts_pos = ts_pos
        self.vert_pos = vert_pos
        self.data_id_pos = data_id_pos

        if from_file is not None:
            self.from_file(from_file)
        else:
            try:
                self.import_default(self.name)
            except Exception:
                pass

    @property
    def info_init(self):
        """Empty dictionary containing init values of infos to be
        extracted from filenames
        """
        return dict(
            year=None, var_name=None, ts_type=None, vert_code="", is_at_stations=False, data_id=""
        )

    def from_file(self, file):
        """Identify convention from a file

        Currently only two conventions (aerocom2 and aerocom3) exist that are
        identified by the delimiter used.

        Parameters
        ----------
        file : str
            file path or file name

        Returns
        -------
        FileConventionRead
            this object (with updated convention)

        Raises
        ------
        FileConventionError
            if convention cannot be identified

        Example
        -------
        >>> from pyaerocom.io import FileConventionRead
        >>> filename = 'aerocom3_CAM5.3-Oslo_AP3-CTRL2016-PD_od550aer_Column_2010_monthly.nc'
        >>> print(FileConventionRead().from_file(filename))
        pyaeorocom FileConventionRead
        name: aerocom3
        file_sep: _
        year_pos: -2
        var_pos: -4
        ts_pos: -1
        """

        if basename(file).count("_") >= 4:
            self.import_default("aerocom3")
        elif basename(file).count(".") >= 4:
            self.import_default("aerocom2")
        else:
            raise FileConventionError(
                f"Could not identify convention from input file {basename(file)}"
            )
        self.check_validity(file)
        return self

    def check_validity(self, file):
        """Check if filename is valid"""
        info = self.get_info_from_file(file)
        year = info["year"]
        if not TsType.valid(info["ts_type"]):
            raise FileConventionError(
                f"Invalid ts_type {info['ts_type']} in filename {basename(file)}"
            )
        elif not (const.MIN_YEAR <= year <= const.MAX_YEAR):
            raise FileConventionError(f"Invalid year {info['year']} in filename {basename(file)}")

    def _info_from_aerocom3(self, file: str) -> dict:
        """Extract info from filename Aerocom 3 convention

        Parameters
        -----------
        file : str
            netcdf file name

        Returns
        -------
        dict
            dictionary containing infos that were extracted from filename
        """
        # init result dictionary
        info = self.info_init

        spl = splitext(basename(file))[0].split(self.file_sep)
        # phase 3 file naming convention
        try:
            info["year"] = int(spl[self.year_pos])
        except Exception:
            raise FileConventionError(
                f"Failed to extract year information from file {basename(file)} "
                f"using file convention Aerocom 3 {self.name}"
            )
        try:
            # include vars for the surface
            if spl[self.vert_pos].lower() in self.AEROCOM3_VERT_INFO["2d"]:
                info["var_name"] = spl[self.var_pos]
            # also include 3d vars that provide station based data
            # and contain the string vmr in this case the variable name has to
            # be slightly changed to the  aerocom phase 2 naming
            elif spl[self.vert_pos].lower() in self.AEROCOM3_VERT_INFO["3d"]:
                if "vmr" in spl[self.var_pos]:
                    info["var_name"] = spl[self.var_pos].replace("vmr", "vmr3d")
                else:
                    info["var_name"] = spl[self.var_pos]
            else:
                raise FileConventionError(
                    f"Invalid file name (Aerocom 3 naming convention).\n{file}\n"
                    f"Invalid string identifier for vertical coordinate: {spl[self.vert_pos]}"
                )
        except Exception as e:
            raise FileConventionError(
                f"Failed to extract variable name from file {basename(file)} "
                f"using file convention {self.name}.\nError: {repr(e)}"
            )
        try:
            info["ts_type"] = spl[self.ts_pos]
        except Exception:
            raise FileConventionError(
                f"Failed to extract ts_type from file {basename(file)} "
                f"using file convention {self.name}"
            )
        try:
            info["vert_code"] = spl[self.vert_pos]
        except Exception:
            raise FileConventionError(
                f"Failed to extract vert_code from file {basename(file)} "
                f"using file convention {self.name}"
            )

        try:
            info["data_id"] = self.file_sep.join(spl[self.data_id_pos : self.var_pos])
        except Exception:
            raise FileConventionError(
                f"Failed to extract model name from file {basename(file)} "
                f"using file convention {self.name}"
            )
        if "atstations" in file.lower():
            info["is_at_stations"] = True
        return info

    def _info_from_aerocom2(self, file: str) -> dict:
        """Extract info from filename Aerocom 2 convention

        Parameters
        -----------
        file : str
            netcdf file name

        Returns
        -------
        dict
            dictionary containing infos that were extracted from filename
        """
        info = self.info_init
        if self.file_sep == ".":
            spl = basename(file).split(self.file_sep)
        else:
            spl = splitext(basename(file))[0].split(self.file_sep)
        try:
            info["year"] = int(spl[self.year_pos])
        except Exception:
            raise FileConventionError(
                f"Failed to extract year information from file {basename(file)} "
                f"using file convention {self.name}"
            )
        try:
            info["var_name"] = spl[self.var_pos]
        except Exception:
            raise FileConventionError(
                f"Failed to extract variable information from file {basename(file)} "
                f"using file convention {self.name}"
            )
        try:
            info["ts_type"] = spl[self.ts_pos]
        except Exception:
            raise FileConventionError(
                f"Failed to extract ts_type from file {basename(file)} "
                f"using file convention {self.name}"
            )

        try:
            info["data_id"] = ".".join(spl[self.data_id_pos : self.ts_pos])
        except Exception:
            raise FileConventionError(
                f"Failed to extract name from file {basename(file)} "
                f"using file convention {self.name}"
            )
        if "atstations" in file.lower():
            raise Exception(
                "Developers: please debug "
                "(file convention Aerocom 2 should not have atstations encoded in file name)"
            )
        return info

    def get_info_from_file(self, file: str) -> dict:
        """Identify convention from a file

        Currently only two conventions (aerocom2 and aerocom3) exist that are
        identified by the delimiter used.

        Parameters
        ----------
        file : str
            file path or file name

        Returns
        -------
        dict
            dictionary containing keys `year, var_name, ts_type` and
            corresponding variables, extracted from the filename

        Raises
        ------
        FileConventionError
            if convention cannot be identified

        Example
        -------
        >>> from pyaerocom.io import FileConventionRead
        >>> filename = 'aerocom3_CAM5.3-Oslo_AP3-CTRL2016-PD_od550aer_Column_2010_monthly.nc'
        >>> conv = FileConventionRead("aerocom3")
        >>> info = conv.get_info_from_file(filename)
        >>> for item in info.items(): print(item)
        ('year', 2010)
        ('var_name', 'od550aer')
        ('ts_type', 'monthly')
        """
        if self.name == "aerocom3":
            return self._info_from_aerocom3(file)
        if self.name == "aerocom2":
            return self._info_from_aerocom2(file)
        raise FileConventionError(f"Unknown {self.name}")

    def string_mask(self, data_id, var, year, ts_type, vert_which=None):
        """Returns mask that can be used to identify files of this convention

        Parameters
        ----------
        data_id : str
            experiment ID (e.g. GISS-MATRIX.A2.CTRL)
        var : str
            variable string ID (e.g. "od550aer")
        year : int
            desired year of observation (e.g. 2012)
        ts_type : str
            string specifying temporal resolution (e.g. "daily")

        Example
        -------

        conf_aero2 = FileConventionRead(name="aerocom2")
        conf_aero3 = FileConventionRead(name="aerocom2")

        var = od550aer
        year = 2012
        ts_type = "daily"

        match_str_aero2 = conf_aero2.string_mask(var, year, ts_type)

        match_str_aero3 = conf_aero3.string_mask(var, year, ts_type)

        """
        if ts_type is None:
            ts_type = "*"
        if self.name == "aerocom2":
            if vert_which is not None:
                raise FileConventionError(
                    f"Specification of vert_which ({vert_which}) is not supported for "
                )

            return ".".join([".*", data_id, ts_type, var, str(year), "nc"])
        elif self.name == "aerocom3":
            if vert_which is None:
                vert_which = ".*"
            return "_".join([".*", data_id, var, vert_which, str(year), ts_type]) + ".nc"
        else:
            raise NotImplementedError(
                f"File matching mask for convention {self.name} not yet defined..."
            )

    def import_default(self, name: str):
        """Checks and load default information from database"""

        conf_reader = ConfigParser()
        with resources.path("pyaerocom.data", "file_conventions.ini") as path:
            conf_reader.read(path)
        if not name in conf_reader:
            raise NameError(f"No default available for {name}")
        self.name = name
        for key, val in conf_reader[name].items():
            if key in self.__dict__:
                try:
                    val = int(val)
                except Exception:
                    pass
                self.__dict__[key] = val

    def from_dict(self, new_vals):
        """Load info from dictionary

        Parameters
        ----------
        new_vals : dict
            dictionary containing information

        Returns
        -------
        self
        """
        for k, v in new_vals.items():
            if k in self.__dict__:
                self.__dict__[k] = v
        return self

    def to_dict(self):
        """Convert this object to ordered dictionary"""
        return dict(
            name=self.name,
            file_sep=self.file_sep,
            year_pos=self.year_pos,
            var_pos=self.var_pos,
            ts_pos=self.ts_pos,
            vert_pos=self.vert_pos,
            data_id_pos=self.data_id_pos,
        )

    def __repr__(self):
        return f"{self.name} {super().__repr__()}"

    def __str__(self):
        s = "\npyaeorocom FileConventionRead"
        for k, v in self.to_dict().items():
            s += f"\n{k}: {v}"
        return s



import glob
import logging
import os
import shutil
from pathlib import Path

from pyaerocom import const
from pyaerocom.aeroval.modelentry import ModelEntry
from pyaerocom.aeroval.varinfo_web import VarinfoWeb
from pyaerocom.colocateddata import ColocatedData
from pyaerocom.colocation_auto import Colocator
from pyaerocom.exceptions import TemporalResolutionError
from pyaerocom.griddeddata import GriddedData
from pyaerocom.helpers import (
    get_highest_resolution,
    get_max_period_range,
    make_dummy_cube,
    start_stop_str,
)
from pyaerocom.io import ReadGridded
from pyaerocom.tstype import TsType

logger = logging.getLogger(__name__)


def check_var_ranges_avail(model_data, var_name):
    """
    Check if lower and upper variable ranges are available for input variable

    Parameters
    ----------
    model_data : GriddedData
        modeldata containing variable data
    var_name : str
        variable name to be checked (must be the same as model data
        AeroCom variable name).

    Raises
    ------
    ValueError
        if ranges for input variable are not defined and if input model data
        corresponds to a different variable than the input variable name.

    Returns
    -------
    None

    """
    try:
        VarinfoWeb(var_name)
    except AttributeError:
        if model_data.var_name_aerocom == var_name:
            model_data.register_var_glob(delete_existing=True)
        else:
            raise ValueError(
                f"Mismatch between variable name of input model_data "
                f"({model_data.var_name_aerocom}) and var_name {var_name}"
            )


def _check_statistics_periods(periods: list) -> list:
    """
    Check input list of period strings is valid

    Parameters
    ----------
    periods : list
        list containing period strings to be checked.

    Raises
    ------
    ValueError
        if input is not a list or any of the provided periods in that list is
        not a string or invalid.

    Returns
    -------
    list
        list of periods

    """
    checked = []
    if not isinstance(periods, list):
        raise ValueError("statistics_periods needs to be a list")
    for per in periods:
        if not isinstance(per, str):
            raise ValueError("All periods need to be strings")
        spl = [x.strip() for x in per.split("-")]
        if len(spl) > 2:
            raise ValueError(
                f"Invalid value for period ({per}), can be either single "
                f"years or period of years (e.g. 2000-2010)."
            )
        _per = "-".join([str(int(val)) for val in spl])
        checked.append(_per)
    return checked


def _period_str_to_timeslice(period: str) -> slice:
    """
    Convert input period to a time slice

    Parameters
    ----------
    period : str
        period, e.g. "2000-2010"

    Raises
    ------
    ValueError
        if input period is invalid

    Returns
    -------
    slice
        slice containing start and end strings.
    """
    spl = period.split("-")
    if len(spl) == 1:
        return slice(spl[0], spl[0])
    elif len(spl) == 2:
        return slice(*spl)
    raise ValueError(period)


def _get_min_max_year_periods(statistics_periods):
    """Get lowest and highest available year from all periods

    Parameters
    ----------
    statistics_periods : list
        list of periods for experiment

    Returns
    -------
    int
        start year
    int
        stop year (may be the same as start year, e.g. if periods suggest
        single year analysis).
    """
    startyr, stopyr = 1e6, -1e6
    for per in statistics_periods:
        sl = _period_str_to_timeslice(per)
        perstart, perstop = int(sl.start), int(sl.stop)
        if perstart < startyr:
            startyr = perstart
        if perstop > stopyr:
            stopyr = perstop
    return startyr, stopyr


def make_dummy_model(obs_list: list, cfg) -> str:

    # Sets up variable for the model register
    tmpdir = const.LOCAL_TMP_DIR
    const.add_data_search_dir(tmpdir)

    model_id = "dummy_model"
    outdir = os.path.join(tmpdir, f"{model_id}/renamed")

    os.makedirs(outdir, exist_ok=True)

    # Finds dates and freq to use, so that all observations are covered
    (start, stop) = get_max_period_range(cfg.time_cfg.periods)
    freq = get_highest_resolution(*cfg.time_cfg.freqs)

    # Loops over variables in obs
    for obs in obs_list:
        for var in cfg.obs_cfg[obs]["obs_vars"]:
            # Create dummy cube

            dummy_cube = make_dummy_cube(var, start_yr=start, stop_yr=stop, freq=freq)

            # Converts cube to GriddedData
            dummy_grid = GriddedData(dummy_cube)

            # Loop over each year
            yr_gen = dummy_grid.split_years()

            for dummy_grid_yr in yr_gen:
                # Add to netcdf
                yr = dummy_grid_yr.years_avail()[0]
                vert_code = cfg.obs_cfg[obs]["obs_vert_type"]

                save_name = dummy_grid_yr.aerocom_savename(model_id, var, vert_code, yr, freq)
                dummy_grid_yr.to_netcdf(outdir, savename=save_name)

    # Add dummy model to cfg
    cfg.model_cfg["dummy"] = ModelEntry(model_id="dummy_model")

    return model_id


def delete_dummy_model(model_id: str) -> None:
    tmpdir = const.LOCAL_TMP_DIR
    const.add_data_search_dir(tmpdir)

    renamed = Path(tmpdir) / f"{model_id}/renamed"
    for path in renamed.glob("*.nc"):
        print(f"Deleting dummy model {path}")
        path.unlink()



import logging
from traceback import format_exc

from pyaerocom import const
from pyaerocom._lowlevel_helpers import BrowseDict, ListOfStrings, StrType
from pyaerocom.exceptions import InitialisationError
from pyaerocom.metastandards import DataSource

logger = logging.getLogger(__name__)


class ObsEntry(BrowseDict):
    """Observation configuration for evaluation (dictionary)

    Note
    ----
    Only :attr:`obs_id` and `obs_vars` are mandatory, the rest is optional.

    Attributes
    ----------
    obs_id : str
        ID of observation network in AeroCom database
        (e.g. 'AeronetSunV3Lev2.daily')
    obs_vars : list
        list of pyaerocom variable names that are supposed to be analysed
        (e.g. ['od550aer', 'ang4487aer'])
    obs_ts_type_read : :obj:`str` or :obj:`dict`, optional
        may be specified to explicitly define the reading frequency of the
        observation data (so far, this does only apply to gridded obsdata such
        as satellites). For ungridded reading, the frequency may be specified
        via :attr:`obs_id`, where applicable (e.g. AeronetSunV3Lev2.daily).
        Can be specified variable specific in form of dictionary.
    obs_vert_type : str, optional
        Aerocom vertical code encoded in the model filenames (only AeroCom 3
        and later).
    obs_aux_requires : dict, optional
        information about required datasets / variables for auxiliary
        variables.
    instr_vert_loc : str, optional
        vertical location code of observation instrument. This is used in
        the aeroval interface for separating different categories of measurements
        such as "ground", "space" or "airborne".
    is_superobs : bool
        if True, this observation is a combination of several others which all
        have to have their own obs config entry.
    only_superobs : bool
        this indicates whether this configuration is only to be used as part
        of a superobs network, and not individually.
    read_opts_ungridded : :obj:`dict`, optional
        dictionary that specifies reading constraints for ungridded reading
        (c.g. :class:`pyaerocom.io.ReadUngridded`).
    """

    SUPPORTED_VERT_CODES = ["Column", "Profile", "Surface"]
    ALT_NAMES_VERT_CODES = dict(ModelLevel="Profile")

    SUPPORTED_VERT_LOCS = DataSource.SUPPORTED_VERT_LOCS

    obs_vars = ListOfStrings()
    obs_vert_type = StrType()

    def __init__(self, **kwargs):

        self.obs_id = ""

        self.obs_vars = []
        self.obs_ts_type_read = None
        self.obs_vert_type = ""
        self.obs_aux_requires = {}
        self.instr_vert_loc = None

        self.is_superobs = False
        self.only_superobs = False

        self.read_opts_ungridded = {}

        self.update(**kwargs)
        self.check_cfg()
        self.check_add_obs()

    def check_add_obs(self):
        """Check if this dataset is an auxiliary post dataset"""
        if len(self.obs_aux_requires) > 0:
            if not self.obs_type == "ungridded":
                raise NotImplementedError(
                    f"Cannot initialise auxiliary setup for {self.obs_id}. "
                    f"Aux obs reading is so far only possible for ungridded observations."
                )
            if not self.obs_id in const.OBS_IDS_UNGRIDDED:
                try:
                    const.add_ungridded_post_dataset(**self)
                except Exception:
                    raise InitialisationError(
                        f"Cannot initialise auxiliary reading setup for {self.obs_id}. "
                        f"Reason:\n{format_exc()}"
                    )

    def get_all_vars(self) -> list:
        """
        Get list of all variables associated with this entry

        Returns
        -------
        list
            DESCRIPTION.

        """
        return self.obs_vars

    def has_var(self, var_name):
        """
        Check if input variable is defined in entry

        Returns
        -------
        bool
            True if entry has variable available, else False
        """
        return True if var_name in self.get_all_vars() else False

    def get_vert_code(self, var):
        """Get vertical code name for obs / var combination"""
        vc = self["obs_vert_type"]
        if isinstance(vc, str):
            val = vc
        elif isinstance(vc, dict) and var in vc:
            val = vc[var]
        else:
            raise ValueError(f"invalid value for obs_vert_type: {vc}")
        if not val in self.SUPPORTED_VERT_CODES:
            raise ValueError(
                f"invalid value for obs_vert_type: {val}. Choose from "
                f"{self.SUPPORTED_VERT_CODES}."
            )
        return val

    def check_cfg(self):
        """Check that minimum required attributes are set and okay"""

        if not self.is_superobs and not isinstance(self.obs_id, (str, dict)):
            raise ValueError(
                f"Invalid value for obs_id: {self.obs_id}. Need str or dict "
                f"or specification of ids and variables via obs_compute_post"
            )
        if isinstance(self.obs_vars, str):
            self.obs_vars = [self.obs_vars]
        elif not isinstance(self.obs_vars, list):
            raise ValueError(f"Invalid input for obs_vars. Need list or str, got: {self.obs_vars}")
        ovt = self.obs_vert_type
        if ovt is None:
            raise ValueError(
                f"obs_vert_type is not defined. Please specify "
                f"using either of the available codes: {self.SUPPORTED_VERT_CODES}. "
                f"It may be specified for all variables (as string) "
                f"or per variable using a dict"
            )
        elif isinstance(ovt, str) and not ovt in self.SUPPORTED_VERT_CODES:
            self.obs_vert_type = self._check_ovt(ovt)
        elif isinstance(self.obs_vert_type, dict):
            for var_name, val in self.obs_vert_type.items():
                if not val in self.SUPPORTED_VERT_CODES:
                    raise ValueError(
                        f"Invalid value for obs_vert_type: {self.obs_vert_type} "
                        f"(variable {var_name}). Supported codes are {self.SUPPORTED_VERT_CODES}."
                    )
        ovl = self.instr_vert_loc
        if isinstance(ovl, str) and not ovl in self.SUPPORTED_VERT_LOCS:
            raise AttributeError(
                f"Invalid value for instr_vert_loc: {ovl} for {self.obs_id}. "
                f"Please choose from: {self.SUPPORTED_VERT_LOCS}"
            )

    def _check_ovt(self, ovt):
        """Check if obs_vert_type string is valid alias

        Parameters
        ----------
        ovt : str
            obs_vert_type string

        Returns
        -------
        str
            valid obs_vert_type

        Raises
        ------
        ValueError
            if `ovt` is invalid
        """
        if ovt in self.ALT_NAMES_VERT_CODES:
            _ovt = self.ALT_NAMES_VERT_CODES[ovt]
            logger.warning(f"Please use {_ovt} for obs_vert_code and not {ovt}")
            return _ovt
        valid = self.SUPPORTED_VERT_CODES + list(self.ALT_NAMES_VERT_CODES)
        raise ValueError(
            f"Invalid value for obs_vert_type: {self.obs_vert_type}. "
            f"Supported codes are {valid}."
        )



import logging
from configparser import ConfigParser
from importlib import resources

import numpy as np

from pyaerocom._lowlevel_helpers import BrowseDict

logger = logging.getLogger(__name__)


class DataSource(BrowseDict):
    """Dict-like object defining a data source

    Attributes
    ----------
    data_id
        name (or ID) of dataset (e.g. AeronetSunV3Lev2.daily)
    dataset_name
        name of dataset (e.g. AERONET)
    data_product
        data product (e.g. SDA, Inv, Sun for Aeronet)
    data_version
        version of data (e.g. 3)
    data_level
        level of data (e.g. 2)
    framework : str
        ID of framework to which data is associated (e.g. ACTRIS, GAW)
    instr_vert_loc : str
        Vertical location of measuring instrument(s).
    revision_date
        last revision date of dataset
    ts_type_src
        sampling frequency as defined in data files (use None if undefined)
    stat_merge_pref_attr : str
        optional, a metadata attribute that is available in data and that
        is used to order the individual stations by relevance in case overlaps
        occur. The associated values of this attribute need to be sortable
        (e.g. revision_date). This is only relevant in case overlaps occur.
    """

    SUPPORTED_VERT_LOCS = ["ground", "space", "airborne"]

    _types = dict(
        dataset_name=str,
        data_product=str,
        data_version=float,
        data_level=float,
        framework=str,
        instr_vert_loc=str,
        ts_type_src=str,
        stat_merge_pref_attr=str,
        revision_date=np.datetime64,
        website=str,
    )

    _ini_file_name = "data_sources.ini"

    def __init__(self, **info):

        self.data_id = None
        self.dataset_name = None
        self.data_product = None
        self.data_version = None
        self.data_level = None

        self.framework = None
        self.instr_vert_loc = None
        self.revision_date = None
        self.website = None

        self.ts_type_src = None

        self.stat_merge_pref_attr = None

        self.update(**info)
        if self.data_id is not None:
            self._parse_source_info_from_ini()

    @property
    def data_dir(self):
        """Directory containing data files"""
        from pyaerocom.io.helpers import get_obsnetwork_dir

        return get_obsnetwork_dir(self.data_id)

    def dataset_str(self):
        s = ""
        if self.dataset_name is not None:
            s += self.dataset_name
            hasv = False
            if self.data_version is not None:
                s += f"(v{self.data_version}"
                hasv = True
            if self.data_level is not None:
                if hasv:
                    s += f", Lev {self.data_level})"
                else:
                    s += f"(Lev {self.data_level})"
            else:
                s += ")"
        else:
            s += self.data_id
        return s

    def load_dataset_info(self):
        """Wrapper for :func:`_parse_source_info_from_ini`"""
        try:
            self._parse_source_info_from_ini()
        except Exception:
            pass

    def _parse_source_info_from_ini(self):
        """Parse source info from ini file"""

        if not resources.is_resource("pyaerocom.data", self._ini_file_name):
            raise OSError(f"File {self._ini_file_name} does not exist")

        parser = ConfigParser()
        with resources.path("pyaerocom.data", self._ini_file_name) as path:
            parser.read(path)
        if self.data_id in parser:
            for k, v in parser[self.data_id].items():
                if k in self._types:
                    self[k] = self._types[k](v)
                else:
                    self[k] = str(v)


class StationMetaData(DataSource):
    """This object defines a standard for station metadata in pyaerocom

    Variable names associated with meta data can vary significantly between
    different conventions (e.g. conventions in modellers community vs.
    observations community).

    Note
    ----
    - This object is a dictionary and can be easily expanded
    - In many cases, only some of the attributes are relevant

    Attributes
    ----------
    filename : str
        name of file (may be full path or only filename)
    station_id : str
        Code or unique ID of station
    station_name :str
        name or ID of a station. Note, that the concept of a station in
        pyaerocom is not necessarily related to a fixed coordinate. A station
        can also be a satellite, ship, or a human walking around and measuring
        something
    instrument_name : str
        name (or ID) of instrument
    PI : str
        principal investigator
    country : str
        string specifying country (or country ID)
    ts_type : str
        frequency of data (e.g. monthly). Note the difference between
        :attr:`ts_type_src` of :class:`DataSource`, which specifies the freq.
        of the original files.
    latitude : float
        latitude coordinate
    longitude : float
        longitude coordinate
    altitude : float
        altitude coordinate

    """

    def __init__(self, **info):

        self.filename = None

        self.station_id = None
        self.station_name = None
        self.instrument_name = None
        self.PI = None

        self.country = None
        self.country_code = None

        self.ts_type = None

        self.latitude = np.nan
        self.longitude = np.nan
        self.altitude = np.nan

        super().__init__(**info)


class AerocomDataID:
    """
    Class representing a model data ID following AeroCom PhaseIII conventions

    The ID must contain 4 substrings with meta parameters:

        <ModelName>-<MeteoConfigSpecifier>_<ExperimentName>-<PerturbationName>

    E.g.

        NorESM2-met2010_CTRL-AP3

    For more information see `AeroCom diagnostics spreadsheet <https://docs.google.com/spreadsheets/d/1NiHLVTDsBo0JEBSnnDECNI2ojUnCVlxuy2PFrsRJW38/edit#gid=1475397852>`__

    This interface can be used to make sure a provided data ID is following
    this convention and to extract the corresponding meta parameters as
    dictionary (:func:`to_dict`) or to create an data_id from the corresponding
    meta parameters :func:`from_dict`.
    """

    DELIM = "_"
    SUBDELIM = "-"
    KEYS = ["model_name", "meteo", "experiment", "perturbation"]

    def __init__(self, data_id=None, **meta_info):

        self._data_id = None
        self._values = None

        if data_id is not None:
            self.data_id = data_id
        elif meta_info:
            self._values_from_dict(meta_info)

    @property
    def data_id(self):
        """
        str
            AeroCom data ID
        """
        return self._data_id

    @data_id.setter
    def data_id(self, val):
        self._values = self._eval_data_id(val)
        self._data_id = val

    @property
    def values(self):
        if self._values is not None:
            return self._values
        raise AttributeError("Meta value list is not set.")

    @values.setter
    def values(self, val):
        if not isinstance(val, list) or not len(val) == len(self.KEYS):
            raise ValueError(f"Invalid input: need list of length {len(self.KEYS)}")
        # this will first create a data_id string from input values and
        # then call setter method to make sure the input is correct.
        self.data_id = self.from_values(val)

    def to_dict(self):
        """Convert data_id to dictionary

        Returns
        -------
        dict
            dictionary with metadata information
        """
        if not len(self._values) == len(self.KEYS):
            self._eval_data_id(self.data_id)
        return dict(zip(self.KEYS, self._values))

    def _values_from_dict(self, meta):
        vals = []
        for key in self.KEYS:
            if not key in meta:
                raise KeyError(f"Missing specification of {key} in input meta dict")
            vals.append(meta[key])
        self._data_id = self.from_values(vals)
        self._values = vals

    @staticmethod
    def from_dict(meta):
        """
        Create instance of AerocomDataID from input meta dictionary

        Parameters
        ----------
        meta : dict
            dictionary containing required keys (cf. :attr:`KEYS`) and
            corresponding values to create an data_id

        Raises
        ------
        KeyError
            if not all information required is provided

        Returns
        -------
        AerocomDataID

        """
        return AerocomDataID(**meta)

    @staticmethod
    def from_values(values):
        """
        Create data_id from list of values

        Note
        ----
        The values have to be in the right order, cf. :attr:`KEYS`

        Parameters
        ----------
        values : list
            list containing values for each key in :attr:`KEYS`

        Raises
        ------
        ValueError
            if length of input list mismatches length of :attr:`KEYS`

        Returns
        -------
        str
            generated data_id

        """
        if not len(values) == 4:
            raise ValueError("Need 4 entries model_name, meteo_config, experiment, perturbation")
        return "{}-{}_{}-{}".format(*values)

    def _eval_data_id(self, val):
        """
        Check and extract meta information from input data_id

        Parameters
        ----------
        val : str
            data_id

        Raises
        ------
        ValueError
            if input is not string or is not in format
            <model_name>-<meteo_config>_<experiment>-<perturbation>

        Returns
        -------
        values
            DESCRIPTION.

        """
        if not isinstance(val, str):
            raise ValueError(f"Invalid input for data_id. Need str. Got {val}")

        values = [""] * len(self.KEYS)
        spl = val.split(self.DELIM)
        if not len(spl) == 2:
            logger.warning(
                f"Invalid data ID {val}. Need format <model-name>_<meteo-config>_<eperiment-name>"
            )
            values[0] = val
            return values

        sub = spl[0].split(self.SUBDELIM, 1)
        if len(sub) == 2:
            values[0] = sub[0]  # model_name

            meteo = sub[1]
            if meteo.startswith("met"):
                values[1] = meteo  # meteo_config
            else:
                logger.warning(
                    f"Meteorology config substring in data_id {meteo} needs to start with met."
                )
                values[0] = spl[0]
        else:
            values[0] = spl[0]

        sub = spl[1].split(self.SUBDELIM, 1)
        if len(sub) == 2:
            values[2] = sub[0]
            values[3] = sub[1]
        else:
            values[2] = spl[1]
        return values

    def __eq__(self, other):
        return True if self._data_id == str(other) else False

    def __repr__(self):
        return self._data_id

    def __str__(self):
        return self._data_id


STANDARD_META_KEYS = list(StationMetaData())



"""
Settings and helper methods / classes for I/O of obervation data

Note
----
Some settings like paths etc can be found in :mod:`pyaerocom.config.py`
"""
from pyaerocom._lowlevel_helpers import dict_to_str, str_underline

#: Wavelength tolerance for observations if data for required wavelength
#: is not available
OBS_WAVELENGTH_TOL_NM = 10.0

#: This boolean can be used to enable / disable the former (i.e. use
#: available wavelengths of variable in a certain range around variable
#: wavelength).
OBS_ALLOW_ALT_WAVELENGTHS = True


class ObsVarCombi:
    def __init__(self, obs_id, var_name):
        self.obs_id = obs_id
        self.var_name = var_name

    def __repr__(self):
        return f"{self.obs_id};{self.var_name}"

    def __str__(self):
        return repr(self)


class AuxInfoUngridded:
    MAX_VARS_PER_METHOD = 2

    def __init__(
        self, data_id, vars_supported, aux_requires, aux_merge_how, aux_funs=None, aux_units=None
    ):

        self.data_id = data_id
        self.vars_supported = vars_supported

        self.aux_requires = aux_requires
        self.aux_merge_how = aux_merge_how
        self.aux_funs = aux_funs
        self.aux_units = aux_units
        self.check_status()

    def to_dict(self):
        """Dictionary representation of this object

        Ignores any potential private attributes.
        """
        dd = {}
        for key, val in self.__dict__.items():
            if any([key.startswith(x) for x in ("_", "__")]):
                continue
            dd[key] = val
        return dd

    def check_status(self):
        """
        Check if specifications are correct and consistent

        Raises
        ------
        ValueError
            If one of the class attributes is invalid
        NotImplementedError
            If computation method contains more than 2 variables / datasets

        """
        if isinstance(self.vars_supported, str):
            self.vars_supported = [self.vars_supported]

        if isinstance(self.aux_merge_how, str):
            nv = len(self.vars_supported)
            self.aux_merge_how = dict(zip(self.vars_supported, [self.aux_merge_how] * nv))
        if self.aux_funs is None:
            self.aux_funs = {}
        if self.aux_units is None:
            self.aux_units = {}

        for var in self.vars_supported:
            if not var in self.aux_requires:
                raise ValueError(f"Variable {var} is not defined in attr aux_requires...")

            elif not var in self.aux_merge_how:
                raise ValueError(
                    f"Missing information about how {var} should be merged (aux_merge_how)"
                )
            merge_how = self.aux_merge_how[var]
            if merge_how == "eval":
                if not var in self.aux_funs:
                    raise ValueError(
                        f"Specification of computation function is missing for var {var}"
                    )
                fun = self.aux_funs[var]

                if not isinstance(fun, str):
                    raise ValueError("eval functions need to be strings")

            aux_info = self.aux_requires[var]

            fc = 0
            for aux_id, var_info in aux_info.items():
                if isinstance(var_info, str):
                    # make sure variables are represented as list, even if
                    # it is only one
                    aux_info[aux_id] = var_info = [var_info]
                for _var in var_info:
                    obsvar = ObsVarCombi(aux_id, _var)
                    obsvarstr = str(obsvar)
                    if merge_how == "eval" and not obsvarstr in fun:
                        raise ValueError(
                            f"Mismatch between aux_requires and aux_funs for variable {var}. "
                            f"No such obs;var string {obsvarstr} in computation method {fun}"
                        )

                    fc += 1
                    if fc > self.MAX_VARS_PER_METHOD:
                        raise NotImplementedError("So far only 2 variables can be combined...")

    def __repr__(self):
        return f"{type(self).__name__}; data_id: {self.data_id}; vars_supported: {self.vars_supported}"

    def __str__(self):
        name = str(type(self).__name__)
        s = str_underline(name)
        s += dict_to_str(self.to_dict())
        return s



import logging
import os
import tarfile
from pathlib import Path
from traceback import format_exc

import requests

from pyaerocom import const
from pyaerocom.io import (
    ReadAeronetInvV3,
    ReadAeronetSdaV3,
    ReadAeronetSunV3,
    ReadAirNow,
    ReadEarlinet,
    ReadEbas,
    ReadEEAAQEREP_V2,
    ReadGhost,
)

logger = logging.getLogger(__name__)


class AccessTestData:
    #: That's were the testdata can be downloaded from
    URL_TESTDATA = (
        "https://pyaerocom-ng.met.no/pyaerocom-suppl/testdata-minimal.tar.gz.ebas_202201"
    )

    #: Directory where testdata will be downloaded into
    BASEDIR_DEFAULT = const.OUTPUTDIR

    #: Name of testdata directory
    TESTDATADIRNAME = "testdata-minimal"

    #: Paths to be added to pya.const. All relative to :attr:`basedir`
    ADD_PATHS = {
        "MODELS": "modeldata",
        "OBSERVATIONS": "obsdata",
        "CONFIG": "config",
        "AeronetSunV3L2Subset.daily": "obsdata/AeronetSunV3Lev2.daily/renamed",
        "AeronetSunV3L2Subset.AP": "obsdata/AeronetSunV3Lev2.0.AP/renamed",
        "AeronetSDAV3L2Subset.daily": "obsdata/AeronetSDAV3Lev2.daily/renamed",
        "AeronetInvV3L2Subset.daily": "obsdata/AeronetInvV3Lev2.daily/renamed",
        "EBASSubset": "obsdata/EBASMultiColumn",
        "AirNowSubset": "obsdata/AirNowSubset",
        "G.EEA.daily.Subset": "obsdata/GHOST/data/EEA_AQ_eReporting/daily",
        "G.EEA.hourly.Subset": "obsdata/GHOST/data/EEA_AQ_eReporting/hourly",
        "G.EBAS.daily.Subset": "obsdata/GHOST/data/EBAS/daily",
        "G.EBAS.hourly.Subset": "obsdata/GHOST/data/EBAS/hourly",
        "EEA_AQeRep.v2.Subset": "obsdata/EEA_AQeRep.v2/renamed",
        "Earlinet-test": "obsdata/Earlinet",
    }

    _UNGRIDDED_READERS = {
        "AeronetSunV3L2Subset.daily": ReadAeronetSunV3,
        "AeronetSunV3L2Subset.AP": ReadAeronetSunV3,
        "AeronetSDAV3L2Subset.daily": ReadAeronetSdaV3,
        "AeronetInvV3L2Subset.daily": ReadAeronetInvV3,
        "EBASSubset": ReadEbas,
        "AirNowSubset": ReadAirNow,
        "G.EEA.daily.Subset": ReadGhost,
        "G.EEA.hourly.Subset": ReadGhost,
        "G.EBAS.daily.Subset": ReadGhost,
        "G.EBAS.hourly.Subset": ReadGhost,
        "EEA_AQeRep.v2.Subset": ReadEEAAQEREP_V2,
        "Earlinet-test": ReadEarlinet,
    }

    def __init__(self, basedir=None):
        self._basedir = None
        if basedir is not None:
            self.basedir = basedir

    @property
    def basedir(self):
        """Directory into which testdata will be downloaded"""
        if self._basedir is not None:
            return Path(self._basedir)
        return Path(self.BASEDIR_DEFAULT)

    @basedir.setter
    def basedir(self, val):
        if not os.path.isdir(val):
            raise ValueError(f"Input basedir {val} is not a directory...")
        self._basedir = val

    @property
    def testdatadir(self):
        """Directory containing testdata"""
        return self.basedir.joinpath(self.TESTDATADIRNAME)

    def download(self, basedir=None):
        """
        Download testdata

        Parameters
        ----------
        basedir : str, optional
            directory in which testdata is supposed to be downloaded
        Returns
        -------
        bool
            True if download was successful, else False

        """
        if basedir is not None:
            self.basedir = basedir
        logger.info(f"Downloading pyaerocom testdata into {self.basedir}")

        download_loc = self.basedir.joinpath(f"{self.TESTDATADIRNAME}.tar.gz")

        try:
            r = requests.get(self.URL_TESTDATA)
            with open(download_loc, "wb") as f:
                f.write(r.content)

            with tarfile.open(download_loc, "r:gz") as tar:
                tar.extractall(const.OUTPUTDIR)
                tar.close()
        except Exception:
            logger.warning(f"Failed to download testdata. Traceback:\n{format_exc()}")
            return False
        finally:
            if download_loc.exists():
                os.remove(download_loc)
        return True

    def check_access(self, add_check_paths=None):
        """
        Method that checks if testdata can be accessed

        See also :func:`check_access_and_download_if_needed`.

        Parameters
        ----------
        add_check_paths : dict, optional
            additional paths used to validate testdata relative to testdata
            directory. Paths in :attr:`ADD_PATHS` will always be validated as
            they are required to be available in testdataset.

        Returns
        -------
        bool
            True if testdata is available and all relevant path locations
            could be validated, else False.

        """
        check_paths = {}
        check_paths.update(self.ADD_PATHS)
        if isinstance(add_check_paths, dict):
            check_paths.update(add_check_paths)

        datadir = self.testdatadir
        if not datadir.exists():
            return False

        for data_id, subdir in check_paths.items():
            if not datadir.joinpath(subdir).exists():
                return False
        return True

    def check_access_and_download_if_needed(self, add_check_paths=None):
        """
        Method that checks if testdata can be accessed and if not downloads it

        Parameters
        ----------
        add_check_paths : dict, optional
            additional paths used to validate testdata relative to testdata
            directory. Paths in :attr:`ADD_PATHS` will always be validated as
            they are required to be available in testdataset.

        Returns
        -------
        bool
            True if testdata is available and all relevant path locations
            could be validated, else False.

        """
        if not self.check_access(add_check_paths):
            try:
                if self.download() and self.check_access(add_check_paths):
                    return True
            except Exception as e:
                logger.warning(f"Failed to access testdata: {e}")
            return False
        return True

    def init(self, add_check_paths=None):
        if not self.check_access_and_download_if_needed(add_check_paths):
            return False

        testdatadir = self.testdatadir
        for name, relpath in self.ADD_PATHS.items():
            ddir = str(testdatadir.joinpath(relpath))
            if name in self._UNGRIDDED_READERS:
                if name in const.OBSLOCS_UNGRIDDED and ddir == const.OBSLOCS_UNGRIDDED[name]:
                    logger.info(f"dataset {name} is already registered")
                    continue
                reader = self._UNGRIDDED_READERS[name]
                try:
                    const.add_ungridded_obs(name, ddir, reader=reader, check_read=False)
                except Exception as e:
                    logger.warning(
                        f"Failed to instantiate testdata since ungridded "
                        f"dataset {name} at {ddir} could not be registered: {e}"
                    )
                    return False
                logger.info(f"Adding ungridded dataset {name} located at {ddir}. Reader: {reader}")

            else:
                const.add_data_search_dir(ddir)
                logger.info(f"Adding data search directory {ddir}.")
        return True


def initialise():
    td = AccessTestData()
    if td.init():
        logger.info(
            f"pyaerocom-testdata is ready to be used. The data "
            f"is available at {td.testdatadir}"
        )
    else:
        logger.warning("Failed to initiate pyaerocom-testdata")



import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import numpy as np
from cartopy.mpl.geoaxes import GeoAxes
from cartopy.mpl.ticker import LatitudeFormatter, LongitudeFormatter
from matplotlib.colors import BoundaryNorm, LogNorm, Normalize
from numpy import ceil, linspace, meshgrid
from pandas import to_datetime

from pyaerocom import const
from pyaerocom._warnings import ignore_warnings
from pyaerocom.exceptions import DataDimensionError
from pyaerocom.mathutils import exponent
from pyaerocom.plot.config import COLOR_THEME, MAP_AXES_ASPECT, ColorTheme
from pyaerocom.plot.helpers import (
    calc_figsize,
    calc_pseudolog_cmaplevels,
    custom_mpl,
    projection_from_str,
)
from pyaerocom.region import Region

MPL_PARAMS = custom_mpl()


def get_cmap_maps_aerocom(color_theme=None, vmin=None, vmax=None):
    """Get colormap using pyAeroCom color scheme

    Parameters
    ----------
    color_theme : :ColorTheme, optional
        instance of pyaerocom color theme. If None, the default schemes is used
    vmin : float, optional
        lower end of value range (only considered for diverging color maps
        with non-symmetric mapping)
    vmax : float, optional
        upper end of value range only considered for diverging color maps
        with non-symmetric mapping)

    Returns
    -------
    colormap
    """
    if color_theme is None:
        color_theme = COLOR_THEME
    if vmin is not None and vmax is not None and vmin < 0 and vmax > 0:
        cmap = plt.get_cmap(color_theme.cmap_map_div)
        if color_theme.cmap_map_div_shifted:
            if not const.GEONUM_AVAILABLE:  # pragma: no cover
                raise ModuleNotFoundError("need geonum to compute shifted colormap")
            from geonum.helpers import shifted_color_map

            cmap = shifted_color_map(vmin, vmax, cmap)
        return cmap
    return plt.get_cmap(color_theme.cmap_map)


def set_map_ticks(ax, xticks=None, yticks=None):
    """Set or update ticks in instance of GeoAxes object (cartopy)

    Parameters
    ----------
    ax : cartopy.GeoAxes
        map axes instance
    xticks : iterable, optional
        ticks of x-axis (longitudes)
    yticks : iterable, optional
        ticks of y-axis (latitudes)

    Returns
    -------
    cartopy.GeoAxes
        modified axes instance
    """
    lonleft, lonright = ax.get_xlim()
    digits = 2 - exponent(lonleft)
    digits = 0 if digits < 0 else digits
    tick_format = f".{digits:d}f"
    if xticks is None:
        num_lonticks = 7 if lonleft == -lonright else 6
        xticks = linspace(lonleft, lonright, num_lonticks)
    if yticks is None:
        latleft, latright = ax.get_ylim()
        num_latticks = 7 if latleft == -latright else 6
        yticks = linspace(latleft, latright, num_latticks)
    ax.set_xticks(xticks, crs=ccrs.PlateCarree())
    ax.set_yticks(yticks, crs=ccrs.PlateCarree())

    lon_formatter = LongitudeFormatter(
        number_format=tick_format, degree_symbol="", dateline_direction_label=True
    )
    lat_formatter = LatitudeFormatter(number_format=tick_format, degree_symbol="")

    ax.xaxis.set_major_formatter(lon_formatter)
    ax.yaxis.set_major_formatter(lat_formatter)

    return ax


def init_map(
    xlim=(-180, 180),
    ylim=(-90, 90),
    figh=8,
    fix_aspect=False,
    xticks=None,
    yticks=None,
    color_theme=COLOR_THEME,
    projection=None,
    title=None,
    gridlines=False,
    fig=None,
    ax=None,
    draw_coastlines=True,
    contains_cbar=False,
):
    """Initalise a map plot

    Parameters
    ----------
    xlim : tuple
        2-element tuple specifying plotted longitude range
    ylim : tuple
        2-element tuple specifying plotted latitude range
    figh : int
        height of figure in inches
    fix_aspect : bool, optional
        if True, the aspect of the GeoAxes instance is kept fix using the
        default aspect ``MAP_AXES_ASPECT`` defined in
        :mod:`pyaerocom.plot.config`
    xticks : iterable, optional
        ticks of x-axis (longitudes)
    yticks : iterable, optional
        ticks of y-axis (latitudes)
    color_theme : ColorTheme
        pyaerocom color theme.
    projection
        projection instance from cartopy.crs module (e.g. PlateCaree). May also
        be string.
    title : str, optional
        title that is supposed to be inserted
    gridlines : bool
        whether or not to add gridlines to the map
    fig : matplotlib.figure.Figure, optional
        instance of matplotlib Figure class. If specified, the former to
        input args (``figh`` and ``fix_aspect``) are ignored. Note that the
        Figure is wiped clean before plotting, so any plotted content will be
        lost
    ax : GeoAxes, optional
        axes in which the map is plotted
    draw_coastlines : bool
        whether or not to draw coastlines
    contains_cbar : bool
        whether or not a colorbar is intended to be added to the figure (
        impacts the aspect ratio of the figure).

    Returns
    -------
    ax : cartopy.mpl.geoaxes.GeoAxes
        axes instance
    """

    if projection is None:
        projection = ccrs.PlateCarree()
    elif isinstance(projection, str):
        projection = projection_from_str(projection)
    elif not isinstance(projection, ccrs.Projection):
        raise ValueError("Input for projection needs to be instance of cartopy.crs.Projection")

    if not isinstance(ax, GeoAxes):
        if fig is None:
            if not fix_aspect:
                figsize = calc_figsize(xlim, ylim)
            else:
                figw = figh * fix_aspect
                figsize = (figw, figh)

            fig = plt.figure(figsize=figsize)
        else:
            fig.clf()
        if contains_cbar:
            ax = fig.add_axes([0.1, 0.12, 0.75, 0.8], projection=projection)
        else:
            ax = fig.add_axes([0.1, 0.12, 0.85, 0.8], projection=projection)

    if fix_aspect:
        ax.set_aspect(MAP_AXES_ASPECT)

    if not isinstance(color_theme, ColorTheme):
        if isinstance(color_theme, str):
            color_theme = ColorTheme(color_theme)
        else:
            color_theme = COLOR_THEME

    if draw_coastlines:
        ax.coastlines(color=color_theme.color_coastline)

    ax.set_xlim(xlim)
    ax.set_ylim(ylim)

    ax = set_map_ticks(ax, xticks, yticks)
    ax.set_xlabel("Longitude")
    ax.set_ylabel("Latitude")

    if title is not None:
        ax.set_title(title)
    if gridlines:
        ax.gridlines()

    return ax


def plot_griddeddata_on_map(
    data,
    lons=None,
    lats=None,
    var_name=None,
    unit=None,
    xlim=(-180, 180),
    ylim=(-90, 90),
    vmin=None,
    vmax=None,
    add_zero=False,
    c_under=None,
    c_over=None,
    log_scale=True,
    discrete_norm=True,
    cbar_levels=None,
    cbar_ticks=None,
    add_cbar=True,
    cmap=None,
    cbar_ticks_sci=False,
    color_theme=None,
    ax=None,
    ax_cbar=None,
    **kwargs,
):
    """Make a plot of gridded data onto a map

    Parameters
    ----------
    data : ndarray
        2D data array
    lons : ndarray
        longitudes of data
    lats : ndarray
        latitudes of data
    var_name : :obj:`str`, optional
        name of variable that is plotted
    xlim : tuple
        2-element tuple specifying plotted longitude range
    ylim : tuple
        2-element tuple specifying plotted latitude range
    vmin : :obj:`float`, optional
        lower value of colorbar range
    vmax : :obj:`float`, optional
        upper value of colorbar range
    add_zero : bool
        if True and vmin is not 0, then, the colorbar is extended down to 0.
        This may be used, e.g. for logarithmic scales that should include 0.
    c_under : :obj:`float`, optional
        colour of data values smaller than ``vmin``
    c_over : :obj:`float`, optional
        colour of data values exceeding ``vmax``
    log_scale : bool
        if True, the value to color mapping is done in a pseudo log scale
        (see :func:`get_cmap_levels_auto` for implementation)
    discrete_norm : bool
        if True, color mapping will be subdivided into discrete intervals
    cbar_levels : iterable, optional
        discrete colorbar levels. Will be computed automatically, if None
        (and applicable)
    cbar_ticks : iterable, optional
        ticks of colorbar levels. Will be computed automatically, if None
        (and applicable)

    Returns
    -------
    fig
        matplotlib figure instance containing plot result. Use
        ``fig.axes[0]`` to access the map axes instance (e.g. to modify the
        title or lon / lat range, etc.)
    """
    if color_theme is None:
        color_theme = COLOR_THEME
    if add_cbar:
        kwargs["contains_cbar"] = True
    if ax is None:
        ax = init_map(xlim, ylim, color_theme=color_theme, **kwargs)
    if not isinstance(ax, GeoAxes):
        raise ValueError("Invalid input for ax, need GeoAxes")
    fig = ax.figure

    from pyaerocom.griddeddata import GriddedData

    if not isinstance(data, GriddedData):
        raise ValueError("need GriddedData")

    if not data.has_latlon_dims:
        raise DataDimensionError("Input data needs to have latitude and longitude dimension")
    if not data.ndim == 2:
        if not data.ndim == 3 or not "time" in data.dimcoord_names:
            raise DataDimensionError(
                "Input data needs to be 2 dimensional "
                "or 3D with time being the 3rd "
                "dimension"
            )
        data.reorder_dimensions_tseries()
        data = data[0]

    lons = data.longitude.points
    lats = data.latitude.points
    data = data.grid.data

    if add_cbar and ax_cbar is None:
        ax_cbar = _add_cbar_axes(ax)

    X, Y = meshgrid(lons, lats)

    bounds = None
    if cbar_levels is not None:  # user provided levels of colorbar explicitely
        if vmin is not None or vmax is not None:
            raise ValueError("Please provide either vmin/vmax OR cbar_levels")
        bounds = list(cbar_levels)
        low, high = bounds[0], bounds[-1]
        if add_zero and low > 0:
            bounds.insert(0, 0)  # insert zero bound
        if cmap is None:
            cmap = get_cmap_maps_aerocom(color_theme, low, high)
        elif isinstance(cmap, str):
            cmap = plt.get_cmap(cmap)
        norm = BoundaryNorm(boundaries=bounds, ncolors=cmap.N, clip=False)
    else:
        with ignore_warnings(RuntimeWarning, "All-NaN axis encountered"):
            dmin = np.nanmin(data)
            dmax = np.nanmax(data)

        if any([np.isnan(x) for x in [dmin, dmax]]):
            raise ValueError("Cannot plot map of data: all values are NaN")
        elif dmin == dmax:
            raise ValueError(f"Minimum value in data equals maximum value: {dmin}")
        if vmin is None:
            vmin = dmin
        else:
            if vmin < 0 and log_scale:
                log_scale = False
        if vmax is None:
            vmax = dmax
        if exponent(vmin) == exponent(vmax):
            log_scale = False
        if log_scale:  # no negative values allowed
            if vmin < 0:
                vmin = data[data > 0].min()
                if (
                    c_under is None
                ):  # special case, set c_under to indicate that there is values below 0
                    c_under = "r"
            if cmap is None:
                cmap = get_cmap_maps_aerocom(color_theme, vmin, vmax)
            elif isinstance(cmap, str):
                cmap = plt.get_cmap(cmap)
            if discrete_norm:
                # to compute upper range of colour range, round up vmax
                exp = float(exponent(vmax) - 1)
                vmax_colors = ceil(vmax / 10**exp) * 10**exp
                bounds = calc_pseudolog_cmaplevels(vmin=vmin, vmax=vmax_colors, add_zero=add_zero)
                norm = BoundaryNorm(boundaries=bounds, ncolors=cmap.N, clip=False)

            else:
                norm = LogNorm(vmin=vmin, vmax=vmax, clip=True)
        else:
            if add_zero and vmin > 0:
                vmin = 0
            if cmap is None:
                cmap = get_cmap_maps_aerocom(color_theme, vmin, vmax)
            elif isinstance(cmap, str):
                cmap = plt.get_cmap(cmap)
            if discrete_norm:
                bounds = np.linspace(vmin, vmax, 10)
                norm = BoundaryNorm(boundaries=bounds, ncolors=cmap.N, clip=False)
            else:
                norm = Normalize(vmin=vmin, vmax=vmax)
    cbar_extend = "neither"
    if c_under is not None:
        cmap = cmap.copy()
        cmap.set_under(c_under)
        cbar_extend = "min"
        if bounds is not None:
            bounds.insert(0, bounds[0] - bounds[1])
    if c_over is not None:
        cmap = cmap.copy()
        cmap.set_over(c_over)
        if bounds is not None:
            bounds.append(bounds[-1] + bounds[-2])
        if cbar_extend == "min":
            cbar_extend = "both"
        else:
            cbar_extend = "max"
    fig.norm = norm
    disp = ax.pcolormesh(X, Y, data, cmap=cmap, norm=norm, shading="auto")

    if add_cbar:
        cbar = fig.colorbar(disp, extend=cbar_extend, cax=ax_cbar, shrink=0.8)
        fig.cbar = cbar

        if var_name is not None:
            var_str = var_name  # + VARS.unit_str
            if unit is not None:
                if not str(unit) in ["1", "no_unit"]:
                    var_str += f" [{unit}]"

            cbar.set_label(var_str)

        if cbar_ticks:
            cbar.set_ticks(cbar_ticks)
        if cbar_ticks_sci:
            lbls = []
            for lbl in cbar.ax.get_yticklabels():
                tstr = lbl.get_text()
                if bool(tstr):
                    lbls.append(f"{float(tstr):.1e}")
                else:
                    lbls.append("")
            cbar.ax.set_yticklabels(lbls)

    return fig


def _add_cbar_axes(ax):  # , where='right'):
    _loc = ax.bbox._bbox
    fig = ax.figure

    ax_cbar = fig.add_axes([_loc.x1 + 0.02, _loc.y0, 0.02, _loc.y1 - _loc.y0])
    return ax_cbar


def plot_map_aerocom(data, region, **kwargs):
    """High level map plotting function for Aerocom default plotting

    Note
    ----
    This function does not iterate over a cube in time, but uses the
    first available time index in the data.

    Parameters
    ----------
    data : GriddedData
        input data from one timestamp (if data contains more than one time
        stamp, the first index is used)
    region : str or Region
        valid region ID or region

    """
    from pyaerocom import GriddedData

    if not isinstance(data, GriddedData):
        raise ValueError(
            "This plotting method needs an instance of pyaerocom "
            "GriddedData on input, got: %s" % type(data)
        )

    if isinstance(region, str):
        region = Region(region)
    elif not isinstance(region, Region):
        raise ValueError("Invalid input for region, need None, str or Region")

    data = data.filter_region(region.region_id)
    s = data.plot_settings

    vmin, vmax, levs = s.map_vmin, s.map_vmax, None
    if isinstance(s.map_cbar_levels, list) and len(s.map_cbar_levels) > 0:
        vmin, vmax = None, None
        levs = s.map_cbar_levels
    fig = plot_griddeddata_on_map(
        data,
        xlim=region.lon_range_plot,
        ylim=region.lat_range_plot,
        vmin=vmin,
        vmax=vmax,
        c_over=s.map_c_over,
        c_under=s.map_c_under,
        cbar_levels=levs,
        xticks=region.lon_ticks,
        yticks=region.lat_ticks,
        cbar_ticks=s.map_cbar_ticks,
        **kwargs,
    )
    ax = fig.axes[0]

    # annotate model in lower left corner
    lonr, latr = region.lon_range_plot, region.lat_range_plot
    ax.annotate(
        data.data_id,
        xy=(lonr[0] + (lonr[1] - lonr[0]) * 0.03, latr[0] + (latr[1] - latr[0]) * 0.03),
        xycoords="data",
        horizontalalignment="left",
        color="black",
        fontsize=MPL_PARAMS["axes.titlesize"] + 2,
        bbox=dict(boxstyle="square", facecolor="white", edgecolor="none", alpha=0.7),
    )
    ax.annotate(
        "source: AEROCOM",
        xy=(0.97, 0.03),
        xycoords="figure fraction",
        horizontalalignment="right",
        fontsize=MPL_PARAMS["xtick.labelsize"],
        bbox=dict(boxstyle="square", facecolor="none", edgecolor="black"),
    )

    var = data.var_name.upper()
    avg = data.mean()
    start = to_datetime(data.start).strftime("%Y%m%d")
    tit = f"{var} {start} mean {avg.round(3)}"
    ax.set_title(tit)
    return fig


def plot_nmb_map_colocateddata(
    coldata,
    in_percent=True,
    vmin=-100,
    vmax=100,
    cmap=None,
    s=80,
    marker=None,
    step_bounds=None,
    add_cbar=True,
    norm=None,
    cbar_extend=None,
    add_mean_edgecolor=True,
    ax=None,
    ax_cbar=None,
    cbar_outline_visible=False,
    cbar_orientation=None,
    ref_label=None,
    stats_area_weighted=False,
    **kwargs,
):
    """Plot map of normalised mean bias from instance of ColocatedData

    Parameters
    ----------
    coldata : ColocatedData
        data object
    in_percent : bool
        plot bias in percent
    vmin : int
        minimum value of colormapping
    vmax : int
        maximum value of colormapping
    cmap : str or cmap
        colormap used, defaults to bwr
    s : int
        size of marker
    marker : str
        marker used
    step_bounds : int, optional
        step used for discrete colormapping (if None, continuous is used)
    cbar_extend : str
        extend colorbar
    ax : GeoAxes, optional
        axes into which the bias is supposed to be plotted
    ax_cbar : plt.Axes, optional
        axes for colorbar
    cbar_outline_visible : bool
        if False, borders of colorbar are removed
    cbar_orientation : str
        e.g. 'vertical', defaults to 'vertical'
    **kwargs
        keyword args passed to :func:`init_map`

    Returns
    -------
    GeoAxes
    """
    if cbar_extend is None:
        cbar_extend = "both"
    if cbar_orientation is None:
        cbar_orientation = "vertical"
    if cmap is None:
        cmap = "bwr"
    if marker is None:
        marker = "o"
    try:
        mec = kwargs.pop("mec")
    except KeyError:
        try:
            mec = kwargs.pop("markeredgecolor")
        except KeyError:
            mec = "face"

    try:
        mew = kwargs.pop("mew")
    except KeyError:
        mew = 1
    if not coldata.ndim in (3, 4):
        raise DataDimensionError("only 3D or 4D colocated data objects are supported")
    assert "time" in coldata.dims

    mean_bias = coldata.calc_nmb_array()

    if mean_bias.ndim == 1:
        (lats, lons, data) = mean_bias.latitude, mean_bias.longitude, mean_bias.data
    elif "latitude" in mean_bias.dims and "longitude" in mean_bias.dims:
        stacked = mean_bias.stack(latlon=["latitude", "longitude"])
        valid = ~stacked.isnull()
        coords = stacked.latlon[valid].values
        lats, lons = list(zip(*list(coords)))
        data = stacked.data[tuple(valid)]

    if ref_label is None:
        ref_label = coldata.metadata["data_source"][0]

    if in_percent:
        data *= 100
    if ax is None:
        ax = init_map(contains_cbar=True, **kwargs)

    if not isinstance(ax, GeoAxes):
        raise TypeError("Input axes need to be instance of cartopy.GeoAxes")

    fig = ax.figure

    if isinstance(cmap, str):
        cmap = plt.get_cmap(cmap)
    if norm is None and step_bounds is not None:
        bounds = np.arange(vmin, vmax + step_bounds, step_bounds)
        norm = BoundaryNorm(boundaries=bounds, ncolors=cmap.N, clip=False)

    if add_mean_edgecolor:
        nn = Normalize(vmin=vmin, vmax=vmax)
        nmb = coldata.calc_statistics(use_area_weights=stats_area_weighted)["nmb"]
        if in_percent:
            nmb *= 100
        ec = cmap(nn(nmb))
    else:
        ec = mec
    _sc = ax.scatter(
        lons,
        lats,
        c=data,
        marker=marker,
        cmap=cmap,
        vmin=vmin,
        vmax=vmax,
        s=s,
        norm=norm,
        label=ref_label,
        edgecolors=ec,
        linewidths=mew,
    )
    if add_cbar:
        if ax_cbar is None:
            ax_cbar = _add_cbar_axes(ax)
        cbar = fig.colorbar(_sc, extend=cbar_extend, cax=ax_cbar, orientation=cbar_orientation)

        cbar.outline.set_visible(cbar_outline_visible)
        cbar.set_label("NMB [%]")

    return ax



"""
read Aeronet direct sun V2 data
"""
import re

import numpy as np
import pandas as pd

from pyaerocom import const
from pyaerocom.aux_var_helpers import calc_ang4487aer, calc_od550aer
from pyaerocom.io.readaeronetbase import ReadAeronetBase
from pyaerocom.stationdata import StationData


class ReadAeronetSunV2(ReadAeronetBase):
    """Interface for reading Aeronet direct sun version 2 Level 2.0 data

    Note
    ----
    Maintenance for AERONET version 2 data is deprecated since pyaerocom
    version 0.11.0. Please use version 3 data and associated pyaerocom readers.
    """

    #: Mask for identifying datafiles
    _FILEMASK = "*.lev20"

    #: version log of this class (for caching)
    __version__ = "0.16_" + ReadAeronetBase.__baseversion__

    #: Name of dataset (OBS_ID)
    DATA_ID = const.AERONET_SUN_V2L2_AOD_DAILY_NAME

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [
        const.AERONET_SUN_V2L2_AOD_DAILY_NAME,
        const.AERONET_SUN_V2L2_AOD_ALL_POINTS_NAME,
    ]

    #: dictionary assigning temporal resolution flags for supported datasets
    #: that are provided in a defined temporal resolution
    TS_TYPES = {const.AERONET_SUN_V2L2_AOD_DAILY_NAME: "daily"}

    #: default variables for read method
    DEFAULT_VARS = ["od550aer"]

    #: value corresponding to invalid measurement
    NAN_VAL = -9999.0

    #: Dictionary that specifies the index for each data column
    COL_INDEX = dict(
        date=0,
        time=1,
        julien_day=2,
        od1640aer=3,
        od1020aer=4,
        od870aer=5,
        od675aer=6,
        od667aer=7,
        od555aer=8,
        od551aer=9,
        od532aer=10,
        od531aer=11,
        od500aer=12,
        od440aer=15,
        od380aer=17,
        od340aer=18,
        ang4487aer=37,
    )

    #: dictionary containing information about additionally required variables
    #: for each auxiliary variable (i.e. each variable that is not provided
    #: by the original data but computed on import)
    AUX_REQUIRES = {
        "od550aer": ["od440aer", "od500aer", "ang4487aer"],
        "ang4487aer_calc": ["od440aer", "od870aer"],
    }

    #: Functions that are used to compute additional variables (i.e. one
    #: for each variable defined in AUX_REQUIRES)
    AUX_FUNS = {"od550aer": calc_od550aer, "ang4487aer_calc": calc_ang4487aer}

    #: List of variables that are provided by this dataset (will be extended
    #: by auxiliary variables on class init, for details see __init__ method of
    #: base class ReadUngriddedBase)
    PROVIDES_VARIABLES = list(COL_INDEX)[3:]

    @property
    def col_index(self):
        """Dictionary that specifies the index for each data column

        Note
        ----
        Overload of method in base class
        """
        return self.COL_INDEX

    def read_file(self, filename, vars_to_retrieve=None, vars_as_series=False):
        """Read Aeronet Sun V2 level 2 file

        Parameters
        ----------
        filename : str
            absolute path to filename to read
        vars_to_retrieve : :obj:`list`, optional
            list of str with variable names to read. If None, use
            :attr:`DEFAULT_VARS`
        vars_as_series : bool
            if True, the data columns of all variables in the result dictionary
            are converted into pandas Series objects

        Returns
        -------
        StationData
            dict-like object containing results

        Example
        -------
        >>> import pyaerocom.io.read_aeronet_sunv2
        >>> obj = pyaerocom.io.read_aeronet_sunv2.ReadAeronetSunV2()
        >>> files = obj.get_file_list()
        >>> filedata = obj.read_file(files[0])
        """
        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        # implemented in base class
        vars_to_read, vars_to_compute = self.check_vars_to_retrieve(vars_to_retrieve)

        # create empty data object (is dictionary with extended functionality)
        data_out = StationData()
        data_out.data_id = self.data_id

        # create empty array for all variables that are supposed to be read
        for var in vars_to_read:
            data_out[var] = []

        # Iterate over the lines of the file
        self.logger.info(f"Reading file {filename}")
        with open(filename) as in_file:
            # added to output
            data_out.head_line = in_file.readline().strip()
            data_out.algorithm = in_file.readline().strip()
            c_dummy = in_file.readline()
            # re.split(r'=|\,',c_dummy)
            i_dummy = iter(re.split(r"=|\,", c_dummy.rstrip()))
            dict_loc = dict(zip(i_dummy, i_dummy))

            data_out["latitude"] = float(dict_loc["lat"])
            data_out["longitude"] = float(dict_loc["long"])
            data_out["altitude"] = float(dict_loc["elev"])
            data_out["station_name"] = dict_loc["Location"]
            data_out["PI"] = dict_loc["PI"]
            data_out["ts_type"] = self.TS_TYPE
            data_out["instrument_name"] = self.INSTRUMENT_NAME

            c_dummy = in_file.readline()

            # data_out.data_header =
            in_file.readline().strip()

            for line in in_file:
                # process line
                dummy_arr = line.split(self.COL_DELIM)

                day, month, year = dummy_arr[self.col_index["date"]].split(":")

                datestring = "-".join([year, month, day])
                datestring = "T".join([datestring, dummy_arr[self.col_index["time"]]])
                # NOTE JGLISS: parsing timezone offset was removed on 22/2/19
                # since it is deprecated in recent numpy versions, for details
                # see https://www.numpy.org/devdocs/reference/arrays.datetime.html#changes-with-numpy-1-11
                # datestring = '+'.join([datestring, '00:00'])

                data_out["dtime"].append(np.datetime64(datestring))

                for var in vars_to_read:
                    val = float(dummy_arr[self.col_index[var]])
                    if val == self.NAN_VAL:
                        val = np.nan
                    data_out[var].append(val)
        data_out["dtime"] = np.asarray(data_out["dtime"])
        for var in vars_to_read:
            data_out[var] = np.asarray(data_out[var])

        data_out = self.compute_additional_vars(data_out, vars_to_compute)

        # TODO: reconsider to skip conversion to Series
        # convert  the vars in vars_to_retrieve to pandas time series
        # and delete the other ones
        if vars_as_series:
            for var in vars_to_read + vars_to_compute:
                if var in vars_to_retrieve:
                    data_out[var] = pd.Series(data_out[var], index=data_out["dtime"])
                else:
                    del data_out[var]

        return data_out



from pyaerocom._lowlevel_helpers import dict_to_str
from pyaerocom.time_config import TS_TYPES


class GridIO:
    """Global I/O settings for gridded data

    This class includes options related to the import of gridded data. This
    includes both options related to file search as well as preprocessing
    options.

    Attributes
    ----------
    FILE_TYPE : str
        file type of data files. Defaults to .nc
    TS_TYPES : list
        list of strings specifying temporal resolution options encrypted in
        file names.
    PERFORM_FMT_CHECKS : bool
        perform formatting checks when reading netcdf data, using metadata
        encoded in filenames (requires that NetCDF file follows a registered
        naming convention)
    DEL_TIME_BOUNDS : bool
        if True, preexisting bounds on time are deleted when grid data is
        loaded. Else, nothing is done. Aerocom default is True
    SHIFT_LONS : bool
        if True, longitudes are shifted to
        -180 <= lon <= 180 when data is loaded (in case they are defined
        0 <= lon <= 360. Aerocom default is True.
    CHECK_TIME_FILENAME : bool
        the times stored in NetCDF files may be wrong or not stored according
        to the CF conventions. If True, the times are checked and if
        :attr:`CORRECT_TIME_FILENAME`, corrected for on data import based what
        is encrypted in the
        file name. In case of Aerocom models, it is ensured that the filename
        contains both the year and the temporal resolution in the filenames
        (for details see :class:`pyaerocom.io.FileConventionRead`).
        Aerocom default is True
    CORRECT_TIME_FILENAME : bool
        if True and time dimension in data is found to be different from
        filename, it is attempted to be corrected
    EQUALISE_METADATA : bool
        if True (and if metadata varies between different NetCDF files that are
        supposed to be merged in time), the metadata in all loaded objects is
        unified based on the metadata of the first grid (otherwise,
        concatenating them in time might not work using the Iris interface).
        This might need to be reviewed and should be used with care if
        specific metadata aspects of individual files need to be accessed.
        Aerocom default is True
    USE_FILECONVENTION : bool
        if True, file names are strictly required to follow one of the file
        naming conventions that can be specified in the file
        `file_conventions.ini <https://github.com/metno/pyaerocom/tree/master/
        pyaerocom/data>`__. Aerocom default is True.
    INCLUDE_SUBDIRS : bool
        if True, search for files is expanded to all subdirecories included in
        data directory. Aerocom default is False.
    INFER_SURFACE_LEVEL : bool
        if True then surface level for 4D gridded data is inferred automatically
        when necessary (e.g. when extracting surface time series from 4D
        gridded data object that does not contain sufficient information about
        vertical dimension)

    """

    UNITS_ALIASES = {"/m": "m-1"}
    _AEROCOM = {
        "FILE_TYPE": ".nc",
        "PERFORM_FMT_CHECKS": True,
        "DEL_TIME_BOUNDS": True,
        "SHIFT_LONS": True,
        "CHECK_TIME_FILENAME": True,
        "CORRECT_TIME_FILENAME": True,
        "CHECK_DIM_COORDS": True,
        "EQUALISE_METADATA": True,
        "INCLUDE_SUBDIRS": False,
    }

    _DEFAULT = {
        "FILE_TYPE": ".nc",
        "PERFORM_FMT_CHECKS": True,
        "DEL_TIME_BOUNDS": True,
        "SHIFT_LONS": True,
        "CHECK_TIME_FILENAME": True,
        "CORRECT_TIME_FILENAME": True,
        "CHECK_DIM_COORDS": True,
        "EQUALISE_METADATA": True,
        "INCLUDE_SUBDIRS": False,
    }

    def __init__(self, **kwargs):
        self.FILE_TYPE = ".nc"
        # it is important to keep them in the order from highest to lowest
        # resolution
        self.TS_TYPES = TS_TYPES

        self.PERFORM_FMT_CHECKS = True

        # delete time bounds if they exist in netCDF files
        self.DEL_TIME_BOUNDS = True
        # shift longitudes to -180 -> 180 repr (if applicable)
        self.SHIFT_LONS = True

        self.CHECK_TIME_FILENAME = True
        self.CORRECT_TIME_FILENAME = True

        self.CHECK_DIM_COORDS = False
        # check and update metadata dictionary on Cube load since
        # iris concatenate of Cubes only works if metadata is equal

        self.EQUALISE_METADATA = True

        self.INCLUDE_SUBDIRS = False

        self.INFER_SURFACE_LEVEL = True

        self.load_default()

    def load_aerocom_default(self):
        self.from_dict(self._AEROCOM)

    def load_default(self):
        self.from_dict(self._DEFAULT)

    def to_dict(self):
        """Convert object to dictionary

        Returns
        -------
        dict
            settings dictionary
        """
        return self.__dict__

    def from_dict(self, dictionary=None, **settings):
        """Import settings from dictionary"""
        if not dictionary:
            dictionary = {}
        dictionary.update(settings)
        for key, val in dictionary.items():
            self[key] = val

    def __setitem__(self, key, value):
        """Set item

        GridIO["<key>"] = value <=> GridIO.<key> = value
        <=> GridIO.__setitem__(<key>, value)

        Raises
        ------
        IOError
            if key is not a valid setting
        """
        if not key in self.__dict__:
            raise OSError("Could not update IO setting: Invalid key")
        self.__dict__[key] = value

    def __getitem__(self, key):
        """Get item using curly brackets

        GridIO["<key>"] => value

        """
        if not key in self.__dict__:
            raise OSError("Invalid attribute")
        return self.__dict__[key]

    def __str__(self):
        head = f"Pyaerocom {type(self).__name__}"
        return "\n{}\n{}\n{}".format(head, len(head) * "-", dict_to_str(self.to_dict()))



"""
Helpers for conversion of ColocatedData to JSON files for web interface.
"""
import logging
import os
from datetime import datetime

import numpy as np
import pandas as pd
import xarray as xr

from pyaerocom._lowlevel_helpers import read_json, write_json
from pyaerocom._warnings import ignore_warnings
from pyaerocom.aeroval.fairmode_stats import fairmode_stats
from pyaerocom.aeroval.helpers import _get_min_max_year_periods, _period_str_to_timeslice
from pyaerocom.colocateddata import ColocatedData
from pyaerocom.exceptions import (
    AeroValConfigError,
    AeroValTrendsError,
    DataCoverageError,
    TemporalResolutionError,
)
from pyaerocom.helpers import start_stop
from pyaerocom.mathutils import _init_stats_dummy, calc_statistics
from pyaerocom.region import Region, find_closest_region_coord, get_all_default_region_ids
from pyaerocom.region_defs import HTAP_REGIONS_DEFAULT, OLD_AEROCOM_REGIONS
from pyaerocom.trends_engine import TrendsEngine
from pyaerocom.trends_helpers import _get_season_from_months
from pyaerocom.tstype import TsType

logger = logging.getLogger(__name__)


def get_heatmap_filename(ts_type):
    return f"glob_stats_{ts_type}.json"


def get_stationfile_name(station_name, obs_name, var_name_web, vert_code):
    """Get name of station timeseries file"""
    return f"{station_name}_{obs_name}-{var_name_web}_{vert_code}.json"


def get_json_mapname(obs_name, var_name_web, model_name, model_var, vert_code):
    """Get name base name of json file"""
    return f"{obs_name}-{var_name_web}_{vert_code}_{model_name}-{model_var}.json"


def _write_stationdata_json(ts_data, out_dir):
    """
    This method writes time series data given in a dictionary to .json files

    Parameters
    ----------
    ts_data : dict
        A dictionary containing all processed time series data.
    out_dir : str or similar
        output directory

    Returns
    -------
    None.

    """
    filename = get_stationfile_name(
        ts_data["station_name"], ts_data["obs_name"], ts_data["var_name_web"], ts_data["vert_code"]
    )

    fp = os.path.join(out_dir, filename)
    if os.path.exists(fp):
        current = read_json(fp)
    else:
        current = {}
    current[ts_data["model_name"]] = ts_data
    write_json(current, fp, ignore_nan=True)


def _write_site_data(ts_objs, dirloc):
    """Write list of station timeseries files to json"""
    for ts_data in ts_objs:
        # writes json file
        _write_stationdata_json(ts_data, dirloc)


def _write_diurnal_week_stationdata_json(ts_data, out_dirs):
    """
    Minor modification of method _write_stationdata_json to allow a further
    level of sub-directories

    Parameters
    ----------
    ts_data : dict
        A dictionary containing all processed time series data.
    out_dirs : list
        list of file paths for writing data to

    Raises
    ------
    Exception
        Raised if opening json file fails

    Returns
    -------
    None.

    """
    filename = get_stationfile_name(
        ts_data["station_name"], ts_data["obs_name"], ts_data["var_name_web"], ts_data["vert_code"]
    )

    fp = os.path.join(out_dirs["ts/diurnal"], filename)
    if os.path.exists(fp):
        current = read_json(fp)
    else:
        current = {}
    current[ts_data["model_name"]] = ts_data
    write_json(current, fp, ignore_nan=True)


def _add_entry_json(
    heatmap_file, result, obs_name, var_name_web, vert_code, model_name, model_var
):
    if os.path.exists(heatmap_file):
        current = read_json(heatmap_file)
    else:
        current = {}
    if not var_name_web in current:
        current[var_name_web] = {}
    ov = current[var_name_web]
    if not obs_name in ov:
        ov[obs_name] = {}
    on = ov[obs_name]
    if not vert_code in on:
        on[vert_code] = {}
    ovc = on[vert_code]
    if not model_name in ovc:
        ovc[model_name] = {}
    mn = ovc[model_name]
    mn[model_var] = result
    write_json(current, heatmap_file, ignore_nan=True)


def _prepare_regions_json_helper(region_ids):
    regborders, regs = {}, {}
    for regid in region_ids:
        reg = Region(regid)
        name = reg.name
        regs[name] = reg
        regborders[name] = rinfo = {}

        latr = reg.lat_range
        lonr = reg.lon_range
        if any(x is None for x in (latr, lonr)):
            raise ValueError(f"Lat / lon range missing for region {regid}")
        rinfo["minLat"] = latr[0]
        rinfo["maxLat"] = latr[1]
        rinfo["minLon"] = lonr[0]
        rinfo["maxLon"] = lonr[1]

    return (regborders, regs)


def _prepare_default_regions_json():
    return _prepare_regions_json_helper(get_all_default_region_ids())


def _prepare_aerocom_regions_json():
    return _prepare_regions_json_helper(OLD_AEROCOM_REGIONS)


def _prepare_htap_regions_json():
    return _prepare_regions_json_helper(HTAP_REGIONS_DEFAULT)


def _prepare_country_regions(region_ids):
    regs = {}
    for regid in region_ids:
        reg = Region(regid)
        name = reg.name
        regs[name] = reg
    return regs


def init_regions_web(coldata, regions_how):
    regborders, regs = {}, {}
    regborders_default, regs_default = _prepare_default_regions_json()
    if regions_how == "default":
        regborders, regs = regborders_default, regs_default
    elif regions_how == "aerocom":
        regborders, regs = _prepare_aerocom_regions_json()
    elif regions_how == "htap":
        regborders["WORLD"] = regborders_default["WORLD"]
        regs["WORLD"] = regs_default["WORLD"]
        add_borders, add_regs = _prepare_htap_regions_json()
        regborders.update(add_borders)
        regs.update(add_regs)
    elif regions_how == "country":
        regborders["WORLD"] = regborders_default["WORLD"]
        regs["WORLD"] = regs_default["WORLD"]
        coldata.check_set_countries(True)
        regborders.update(coldata.get_country_codes())
        add_regs = _prepare_country_regions(coldata.get_country_codes().keys())
        regs.update(add_regs)
    else:
        raise ValueError("Invalid input for regions_how", regions_how)

    regnames = {}
    for regname, reg in regs.items():
        regnames[reg.region_id] = regname
    return (regborders, regs, regnames)


def update_regions_json(region_defs, regions_json):
    """Check current regions.json for experiment and update if needed

    Parameters
    ----------
    region_defs : dict
        keys are names of region (not IDs!) values define rectangular borders
    regions_json : str
        regions.json file (if it does not exist it will be created).

    Returns
    -------
    dict
        current content of updated regions.json
    """
    if os.path.exists(regions_json):
        current = read_json(regions_json)
    else:
        current = {}

    for region_name, region_info in region_defs.items():
        if not region_name in current:
            current[region_name] = region_info
    write_json(current, regions_json)
    return current


def _init_meta_glob(coldata, **kwargs):
    meta = coldata.metadata

    # create metadata dictionary that is shared among all timeseries files
    meta_glob = {}
    NDSTR = "UNDEFINED"
    try:
        meta_glob["pyaerocom_version"] = meta["pyaerocom"]
    except KeyError:
        meta_glob["pyaerocom_version"] = NDSTR
    try:
        meta_glob["obs_var"] = meta["var_name"][0]
        meta_glob["mod_var"] = meta["var_name"][1]
    except KeyError:
        meta_glob["obs_var"] = NDSTR
        meta_glob["mod_var"] = NDSTR
    try:
        meta_glob["obs_unit"] = meta["var_units"][0]
        meta_glob["mod_unit"] = meta["var_units"][1]
    except KeyError:
        meta_glob["obs_unit"] = NDSTR
        meta_glob["mod_unit"] = NDSTR
    try:
        meta_glob["obs_freq_src"] = meta["ts_type_src"][0]
        meta_glob["mod_freq_src"] = meta["ts_type_src"][1]
    except KeyError:
        meta_glob["obs_freq_src"] = NDSTR
        meta_glob["mod_freq_src"] = NDSTR
    try:
        meta_glob["obs_revision"] = meta["revision_ref"]
    except KeyError:
        meta_glob["obs_revision"] = NDSTR
    meta_glob["processed_utc"] = datetime.utcnow().strftime("%Y-%m-%d %H:%M")
    meta_glob.update(kwargs)
    return meta_glob


def _init_ts_data(freqs):
    data = {}
    for freq in freqs:
        data[f"{freq}_date"] = []
        data[f"{freq}_obs"] = []
        data[f"{freq}_mod"] = []
    return data


def _create_diurnal_weekly_data_object(coldata, resolution):
    """
    Private helper functions that creates the data set containing all the
    weekly time series at the specified resolution. Called by
    _process_sites_weekly_ts. The returned xarray.Dataset contains a dummy time
    variable (currently not used) and the weekly time series as xarray.DataArray
    objects.

    Parameters
    ----------
    coldata : ColocatedData
        ColocatedData object colocated on hourly resolution.
    resolution : string
        String specifying the averaging window used to generate representative
        weekly time series with hourly resolution. Valid values are 'yearly'
        and  'seasonal'.

    Raises
    ------
    ValueError
        If an invalid resolution is given raise ValueError and print what
        was given.

    Returns
    -------
    rep_week_full_period : xarray.Dataset
        Contains the weekly time series as the variable 'rep_week'


    """
    import xarray as xr

    data_allyears = coldata.data

    yearkeys = list(data_allyears.groupby("time.year").groups)

    if resolution == "seasonal":
        seasons = ["DJF", "MAM", "JJA", "SON"]
    elif resolution == "yearly":
        seasons = ["year"]
    else:
        raise ValueError(f"Invalid resolution. Got {resolution}.")

    first = True
    for yr in yearkeys:
        data = data_allyears.where(data_allyears["time.year"] == yr, drop=True)
        for seas in seasons:
            rep_week_ds = xr.Dataset()
            if resolution == "seasonal":
                mon_slice = data.where(data["time.season"] == seas, drop=True)
            elif resolution == "yearly":
                mon_slice = data

            month_stamp = f"{seas}"

            for day in range(7):
                day_slice = mon_slice.where(mon_slice["time.dayofweek"] == day, drop=True)
                rep_day = day_slice.groupby("time.hour").mean(dim="time")
                rep_day["hour"] = rep_day.hour / 24 + day + 1
                if day == 0:
                    rep_week = rep_day
                else:
                    rep_week = xr.concat([rep_week, rep_day], dim="hour")

            rep_week = rep_week.rename({"hour": "dummy_time"})
            month_stamps = np.zeros(rep_week.dummy_time.shape, dtype="<U5")
            month_stamps[:] = month_stamp
            rep_week_ds["rep_week"] = rep_week
            rep_week_ds["month_stamp"] = (("dummy_time"), month_stamps)

            if seas in ["DJF", "year"]:
                rep_week_full_period = rep_week_ds
            else:
                rep_week_full_period = xr.concat([rep_week_full_period, rep_week_ds], dim="period")

        if first:
            output_array = rep_week_full_period
            # output_array = output_array.expand_dims({'year':[0]},0)
            first = False
        else:
            output_array = xr.concat([output_array, rep_week_full_period], dim="year")

    try:
        output_array.year
    except AttributeError:
        output_array = output_array.expand_dims({"year": [0]}, 0)

    output_array = output_array.assign(dict(year=np.array(yearkeys)))
    # output_array.year[:] = np.array(yearkeys)

    return output_array


def _get_period_keys(resolution):
    if resolution == "seasonal":
        period_keys = ["DJF", "MAM", "JJA", "SON"]
    elif resolution == "yearly":
        period_keys = ["Annual"]
    return period_keys


def _process_one_station_weekly(stat_name, i, repw_res, meta_glob, time):
    """
    Processes one station data set for all supported averaging windows into a
    dict of station time series data and metadata.

    Parameters
    ----------
    stat_name : string
        Name of station to process
    i : int
        Index of the station to process in the xarray.Datasets.
    repw_res : dict
        Dictionary of xarray.Datasets for each supported averaging window for
        the weekly time series
    meta_glob : TYPE
        Dictionary containing global metadata.
    time : list
        Time index.

    Returns
    -------
    ts_data : dict
        Dictinary of time series data and metadata for one station. Contains all
        resolutions/averaging windows.
    has_data : bool
        Set to false if all data is missing for a station.

    """
    has_data = False

    years = list(repw_res["seasonal"].year.values)
    yeardict = {}
    for year in years:
        yeardict[f"{year}"] = {}

    ts_data = {
        "time": time,
        "seasonal": {"obs": yeardict, "mod": yeardict},
        "yearly": {"obs": yeardict, "mod": yeardict},
    }
    ts_data["station_name"] = stat_name
    ts_data.update(meta_glob)

    for y, year in enumerate(years):
        for res, repw in repw_res.items():
            repw = repw.transpose("year", "period", "data_source", "dummy_time", "station_name")
            obs_vals = repw[y, :, 0, :, i]
            if (np.isnan(obs_vals)).all().values:
                continue
            has_data = True
            mod_vals = repw[y, :, 1, :, i]

            period_keys = _get_period_keys(res)
            for period_num, pk in enumerate(period_keys):
                ts_data[res]["obs"][f"{year}"][pk] = obs_vals.sel(
                    period=period_num
                ).values.tolist()
                ts_data[res]["mod"][f"{year}"][pk] = mod_vals.sel(
                    period=period_num
                ).values.tolist()
    return ts_data, has_data


def _process_weekly_object_to_station_time_series(repw_res, meta_glob):
    """
    Process the xarray.Datasets objects returned by _create_diurnal_weekly_data_object
    into a dictionary containing station time series data and metadata.

    Parameters
    ----------
    repw_res : dict
        Dictionary of xarray.Datasets for each supported averaging window for
        the weekly time series
    meta_glob : dict
        Dictionary containing global metadata.

    Returns
    -------
    ts_objs : list
        List of dicts containing station time series data and metadata.

    """
    ts_objs = []
    dc = 0
    time = (np.arange(168) / 24 + 1).round(4).tolist()
    for i, stat_name in enumerate(repw_res["seasonal"].station_name.values):
        ts_data, has_data = _process_one_station_weekly(stat_name, i, repw_res, meta_glob, time)

        if has_data:
            ts_objs.append(ts_data)
            dc += 1
    return ts_objs


def _process_weekly_object_to_country_time_series(repw_res, meta_glob, regions_how, region_ids):
    """
    Process the xarray.Dataset objects returned by _create_diurnal_weekly_data_object
    into a dictionary containing country average time series data and metadata.

    ToDo
    ----
    Implement regions_how for other values than country based

    Parameters
    ----------
    repw_res : dict
        Dictionary of xarray.Datasets for each supported averaging window for
        the weekly time series
    meta_glob : dict
        Dictionary containing global metadata.
    regions_how : string
        String describing how regions are to be processed. Regional time series
        are only calculated if regions_how = country.
    region_ids : dict
        Dict containing mapping of region IDs and names.

    Returns
    -------
    ts_objs_reg : list
        List of dicts containing station time series data and metadata.

    """
    ts_objs_reg = []
    time = (np.arange(168) / 24 + 1).round(4).tolist()

    years = list(repw_res["seasonal"].year.values)
    yeardict = {}
    for year in years:
        yeardict[f"{year}"] = {}

    if regions_how != "country":
        print("Regional diurnal cycles are only implemented for country regions, skipping...")
        ts_objs_reg = None
    else:
        for regid, regname in region_ids.items():
            ts_data = {
                "time": time,
                "seasonal": {"obs": yeardict, "mod": yeardict},
                "yearly": {"obs": yeardict, "mod": yeardict},
            }
            ts_data["station_name"] = regname
            ts_data.update(meta_glob)

            for y, year in enumerate(years):
                for res, repw in repw_res.items():
                    repw = repw.transpose(
                        "year", "period", "data_source", "dummy_time", "station_name"
                    )
                    if regid == "WORLD":
                        subset = repw
                    else:
                        subset = repw.where(repw.country == regid)

                    avg = subset.mean(dim="station_name")
                    obs_vals = avg[y, :, 0, :]
                    mod_vals = avg[y, :, 1, :]

                    period_keys = _get_period_keys(res)
                    for period_num, pk in enumerate(period_keys):
                        ts_data[res]["obs"][f"{year}"][pk] = obs_vals.sel(
                            period=period_num
                        ).values.tolist()
                        ts_data[res]["mod"][f"{year}"][pk] = mod_vals.sel(
                            period=period_num
                        ).values.tolist()

            ts_objs_reg.append(ts_data)
    return ts_objs_reg


def _process_sites_weekly_ts(coldata, regions_how, region_ids, meta_glob):
    """
    Private helper function to process ColocatedData objects into dictionaries
    containing represenative weekly time series with hourly resolution.

    Processing the coloceted data object into a collection of representative
    weekly time series is done in the private function _create_diurnal_weekly_data_object.
    This object (an xarray.Dataset) is then further processed into two dictionaries
    containing station and regional time series respectively.

    Parameters
    ----------
    coldata : ColocatedData
        The colocated data to process.
    regions_how : string
        Srting describing how regions are to be processed. Regional time series
        are only calculated if regions_how = country.
    region_ids : dict
        Dict containing mapping of region IDs and names.
    meta_glob : dict
        Dictionary containing global metadata.

    Returns
    -------
    ts_objs : list
        List of dicts containing station time series data and metadata.
    ts_objs_reg : list
        List of dicts containing country time series data and metadata.

    """
    assert coldata.dims == ("data_source", "time", "station_name")

    repw_res = {
        "seasonal": _create_diurnal_weekly_data_object(coldata, "seasonal")["rep_week"],
        "yearly": _create_diurnal_weekly_data_object(coldata, "yearly")["rep_week"].expand_dims(
            "period", axis=0
        ),
    }

    ts_objs = _process_weekly_object_to_station_time_series(repw_res, meta_glob)
    ts_objs_reg = _process_weekly_object_to_country_time_series(
        repw_res, meta_glob, regions_how, region_ids
    )

    return (ts_objs, ts_objs_reg)


def _init_site_coord_arrays(data):
    found = False
    jsdates = {}
    for freq, cd in data.items():
        if cd is None:
            continue
        elif not found:
            sites = cd.data.station_name.values
            lats = cd.data.latitude.values.astype(np.float64)
            lons = cd.data.longitude.values.astype(np.float64)
            if "altitude" in cd.data.coords:
                alts = cd.data.altitude.values.astype(np.float64)
            else:
                alts = [np.nan] * len(lats)
            if "country" in cd.data.coords:
                countries = cd.data.country.values
            else:
                countries = ["UNAVAIL"] * len(lats)

            found = True
        else:
            assert all(cd.data.station_name.values == sites)
        jsdates[freq] = cd.data.jsdate.values.tolist()
    return (sites, lats, lons, alts, countries, jsdates)


def _get_stat_regions(lats, lons, regions):
    regs = []
    for (lat, lon) in zip(lats, lons):
        reg = find_closest_region_coord(lat, lon, regions=regions)
        regs.append(reg)
    return regs


def _process_sites(data, regions, regions_how, meta_glob):
    freqs = list(data)
    (sites, lats, lons, alts, countries, jsdates) = _init_site_coord_arrays(data)
    if regions_how == "country":
        regs = countries
    else:
        regs = _get_stat_regions(lats, lons, regions)

    ts_objs = []
    site_indices = []
    map_meta = []

    for i, site in enumerate(sites):
        # init empty timeseries data object
        site_meta = {
            "station_name": str(site),
            "latitude": lats[i],
            "longitude": lons[i],
            "altitude": alts[i],
        }
        if regions_how == "country":
            site_meta["region"] = [regs[i]]
        else:
            site_meta["region"] = regs[i]
        ts_data = _init_ts_data(freqs)
        ts_data.update(meta_glob)
        ts_data.update(site_meta)
        has_data = False
        for freq, cd in data.items():
            if cd is not None:
                assert cd.dims == ("data_source", "time", "station_name")
                sitedata = cd.data.data[:, :, i]
                if np.all(np.isnan(sitedata)):
                    # skip this site, all is NaN
                    continue
                ts_data[f"{freq}_date"] = jsdates[freq]
                ts_data[f"{freq}_obs"] = sitedata[0].tolist()
                ts_data[f"{freq}_mod"] = sitedata[1].tolist()
                has_data = True
        if has_data:  # site is valid
            # register ts_data
            ts_objs.append(ts_data)
            # remember site indices in data for faster processing of statistics
            # below.
            site_indices.append(i)
            # init map data for each valid site
            map_meta.append({**site_meta})

    return (ts_objs, map_meta, site_indices)


def _get_statistics(obs_vals, mod_vals, min_num):
    stats = calc_statistics(mod_vals, obs_vals, min_num_valid=min_num)
    return _prep_stats_json(stats)


def _make_trends_from_timeseries(obs, mod, freq, season, start, stop, min_yrs):
    """
    Function for generating trends from timeseries

    Includes fomatting in a way
    that can be serialized to json. A key, map_var, is added
    for use in the web interface.

    Parameters
    ----------
    obs     : pd.Series
        Time series of the obs
    mod     : pd.Series
        Time series of the mod
    freq    : str
        Frequency for the trends, either monthly or yearly
    season  : str
        Seasons used for the trends
    start   : int
        Start year
    stop    : int
        Stop year
    min_yrs : int
        Minimal number of years for the calculation of the trends

    Raises
    ------
    AeroValTrendsError
        If stop - start is smaller than min_yrs

    AeroValError
        If the trend engine returns None

    Returns
    ------
    (dict, dict)
        Dicts consiting of the trends data for the obs and mod
    """

    if stop - start < min_yrs:
        raise AeroValTrendsError(f"min_yrs ({min_yrs}) larger than time between start and stop")

    te = TrendsEngine

    # The model and observation data are made to pandas times series
    obs_trend_series = obs
    mod_trend_series = mod

    # Translate season to names used in trends_helpers.py. Should be handled there instead!
    season = _get_season_from_months(season)

    # Trends are calculated
    obs_trend = te.compute_trend(obs_trend_series, freq, start, stop, min_yrs, season)
    mod_trend = te.compute_trend(mod_trend_series, freq, start, stop, min_yrs, season)

    # Makes pd.Series serializable
    if obs_trend["data"] is None or mod_trend["data"] is None:
        raise AeroValTrendsError("Trends came back as None", obs_trend["data"], mod_trend["data"])

    obs_trend["data"] = obs_trend["data"].to_json()
    mod_trend["data"] = mod_trend["data"].to_json()

    obs_trend["map_var"] = f"slp_{start}"
    mod_trend["map_var"] = f"slp_{start}"

    return obs_trend, mod_trend


def _make_trends(obs_vals, mod_vals, time, freq, season, start, stop, min_yrs):
    """
    Function for generating trends from lists of observations

    This will calculate pandas time series
    from the lists and use that to calculate trends

    Parameters
    ----------
    obs     : list
        Time series of the obs
    mod     : list
        Time series of the mod
    freq    : str
        Frequency for the trends, either monthly or yearly
    season  : str
        Seasons used for the trends
    start   : int
        Start year
    stop    : int
        Stop year
    min_yrs : int
        Minimal number of years for the calculation of the trends

    Raises
    ------
    AeroValTrendsError
        If stop - start is smaller than min_yrs

    AeroValError
        If the trend engine returns None

    Returns
    ------
    (dict, dict)
        Dicts consiting of the trends data for the obs and mod
    """

    # The model and observation data are made to pandas times series
    obs_trend_series = pd.Series(obs_vals, time)
    mod_trend_series = pd.Series(mod_vals, time)

    (obs_trend, mod_trend) = _make_trends_from_timeseries(
        obs_trend_series, mod_trend_series, freq, season, start, stop, min_yrs
    )

    return obs_trend, mod_trend


def _process_map_and_scat(
    data,
    map_data,
    site_indices,
    periods,
    main_freq,
    min_num,
    seasons,
    add_trends,
    trends_min_yrs,
    use_fairmode,
    obs_var,
):
    stats_dummy = _init_stats_dummy()
    scat_data = {}
    scat_dummy = [np.nan]
    for freq, cd in data.items():
        for per in periods:
            for season in seasons:
                use_dummy = cd is None
                if not use_dummy:
                    try:
                        subset = _select_period_season_coldata(cd, per, season)
                        jsdate = subset.data.jsdate.values.tolist()
                    except (DataCoverageError, TemporalResolutionError):
                        use_dummy = True
                for i, map_stat in zip(site_indices, map_data):
                    if not freq in map_stat:
                        map_stat[freq] = {}

                    if use_dummy:
                        stats = stats_dummy
                    else:
                        obs_vals = subset.data.data[0, :, i]
                        mod_vals = subset.data.data[1, :, i]
                        stats = _get_statistics(obs_vals, mod_vals, min_num)

                        if use_fairmode and freq != "yearly" and not np.isnan(obs_vals).all():
                            stats["mb"] = np.nanmean(mod_vals - obs_vals)

                            stats["fairmode"] = fairmode_stats(obs_var, stats)

                        #  Code for the calculation of trends
                        if add_trends and freq != "daily":

                            (start, stop) = _get_min_max_year_periods([per])

                            if stop - start >= trends_min_yrs:

                                try:
                                    time = subset.data.time.values
                                    (obs_trend, mod_trend) = _make_trends(
                                        obs_vals,
                                        mod_vals,
                                        time,
                                        freq,
                                        season,
                                        start,
                                        stop,
                                        trends_min_yrs,
                                    )

                                    # The whole trends dicts are placed in the stats dict
                                    stats["obs_trend"] = obs_trend
                                    stats["mod_trend"] = mod_trend

                                except AeroValTrendsError as e:
                                    msg = f"Failed to calculate trends, and will skip. This was due to {e}"
                                    logger.warning(msg)

                    perstr = f"{per}-{season}"
                    map_stat[freq][perstr] = stats
                    if freq == main_freq:
                        # add only sites to scatter data that have data available
                        # in the lowest of the input resolutions (e.g. yearly)
                        site = map_stat["station_name"]
                        if not site in scat_data:
                            scat_data[site] = {}
                            scat_data[site]["latitude"] = map_stat["latitude"]
                            scat_data[site]["longitude"] = map_stat["longitude"]
                            scat_data[site]["altitude"] = map_stat["altitude"]
                            scat_data[site]["region"] = map_stat["region"]
                        if use_dummy:
                            obs = mod = jsdate = scat_dummy
                        else:
                            obs, mod = obs_vals.tolist(), mod_vals.tolist()
                        scat_data[site][perstr] = {"obs": obs, "mod": mod, "date": jsdate}

    return (map_data, scat_data)


def _process_regional_timeseries(data, region_ids, regions_how, meta_glob):
    ts_objs = []
    freqs = list(data)
    check_countries = True if regions_how == "country" else False
    for regid, regname in region_ids.items():
        ts_data = _init_ts_data(freqs)
        ts_data["station_name"] = regname
        ts_data.update(meta_glob)

        for freq, cd in data.items():
            if cd is None:
                continue
            jsfreq = cd.data.jsdate.values.tolist()
            try:
                subset = cd.filter_region(regid, inplace=False, check_country_meta=check_countries)
            except DataCoverageError:
                logger.info(f"no data in {regid} ({freq}) to compute regional timeseries")
                ts_data[f"{freq}_date"] = jsfreq
                ts_data[f"{freq}_obs"] = [np.nan] * len(jsfreq)
                ts_data[f"{freq}_mod"] = [np.nan] * len(jsfreq)
                continue

            if subset.has_latlon_dims:
                avg = subset.data.mean(dim=("latitude", "longitude"))
            else:
                avg = subset.data.mean(dim="station_name")
            obs_vals = avg[0].data.tolist()
            mod_vals = avg[1].data.tolist()
            ts_data[f"{freq}_date"] = jsfreq
            ts_data[f"{freq}_obs"] = obs_vals
            ts_data[f"{freq}_mod"] = mod_vals

        ts_objs.append(ts_data)
    return ts_objs


def _apply_annual_constraint_helper(coldata, yearly):
    arr_yr = yearly.data
    yrs_cd = coldata.data.time.dt.year
    yrs_avail = arr_yr.time.dt.year
    obs_allyrs = arr_yr[0]
    for i, yr in enumerate(yrs_avail):
        obs_yr = obs_allyrs[i]
        nan_sites_yr = obs_yr.isnull()
        if not nan_sites_yr.any():
            continue
        scond = nan_sites_yr.data
        tcond = (yrs_cd == yr).data
        # workaround since numpy sometimes throws IndexError if tcond and
        # scond are attempted to be applied directly via
        # coldata.data.data[:,tcond, scond] = np.nan
        tsel = coldata.data.data[:, tcond]
        tsel[:, :, scond] = np.nan
        coldata.data.data[:, tcond] = tsel

    return coldata


def _apply_annual_constraint(data):
    """
    Apply annual filter to data

    Parameters
    ----------
    data : dict
        keys are frequencies, values are corresponding
        instances of `ColocatedData` in that resolution, or None, if colocated
        data is not available in that resolution (see also
        :func:`_init_data_default_frequencies`).

    Raises
    ------
    AeroValConfigError
        If colocated data in yearly resolution is not available in input data

    Returns
    -------
    output : dict
        like input `data` but with annual constrained applied.

    """
    output = {}
    if not "yearly" in data or data["yearly"] is None:
        raise AeroValConfigError(
            "Cannot apply annual_stats_constrained option. "
            'Please add "yearly" in your setup (see attribute '
            '"statistics_json" in AerocomEvaluation class)'
        )
    yearly = data["yearly"]
    for tst, cd in data.items():
        if cd is None:
            output[tst] = None
        else:
            output[tst] = _apply_annual_constraint_helper(cd, yearly)
    return output


def _prep_stats_json(stats):
    for k, v in stats.items():
        try:
            stats[k] = np.float64(v)  # for json encoder...
        except Exception:
            # value is str (e.g. for weighted stats)
            # 'NOTE': 'Weights were not applied to FGE and kendall and spearman corr (not implemented)'
            stats[k] = v
    return stats


def _get_extended_stats(coldata, use_weights):
    stats = coldata.calc_statistics(use_area_weights=use_weights)

    # Removes the spatial median and temporal mean (see mails between Hilde, Jonas, Augustin and Daniel from 27.09.21)
    # (stats['R_spatial_mean'],
    #  stats['R_spatial_median']) = _calc_spatial_corr(coldata, use_weights)

    # (stats['R_temporal_mean'],
    #  stats['R_temporal_median']) = _calc_temporal_corr(coldata)

    (stats["R_spatial_mean"], _) = _calc_spatial_corr(coldata, use_weights)

    (_, stats["R_temporal_median"]) = _calc_temporal_corr(coldata)

    return _prep_stats_json(stats)


def _calc_spatial_corr(coldata, use_weights):
    """
    Compute spatial correlation both for median and mean aggregation

    Parameters
    ----------
    coldata : ColocatedData
        Input data.
    use_weights : bool
        Apply area weights or not.

    Returns
    -------
    float
        mean spatial correlation
    float
        median spatial correlation

    """
    return (
        coldata.calc_spatial_statistics(aggr="mean", use_area_weights=use_weights)["R"],
        coldata.calc_spatial_statistics(aggr="median", use_area_weights=use_weights)["R"],
    )


def _calc_temporal_corr(coldata):
    """
    Compute temporal correlation both for median and mean aggregation

    Parameters
    ----------
    coldata : ColocatedData
        Input data.

    Returns
    -------
    float
        mean temporal correlation
    float
        median temporal correlation

    """
    if len(coldata.time) < 3:
        return np.nan, np.nan
    elif coldata.has_latlon_dims:
        coldata = coldata.flatten_latlondim_station_name()
    arr = coldata.data
    # Use only sites that contain at least 3 valid data points (otherwise
    # correlation will be 1).
    obs_ok = arr[0].count(dim="time") > 2
    arr = arr.where(obs_ok, drop=True)
    if np.prod(arr.shape) == 0:
        return np.nan, np.nan
    corr_time = xr.corr(arr[1], arr[0], dim="time")
    with ignore_warnings(RuntimeWarning, "Mean of empty slice", "All-NaN slice encountered"):
        return (np.nanmean(corr_time.data), np.nanmedian(corr_time.data))


def _select_period_season_coldata(coldata, period, season):
    tslice = _period_str_to_timeslice(period)
    # expensive, try use solution with numpy indexing directly...
    # also, keep an eye on: https://github.com/pydata/xarray/issues/2799
    arr = coldata.data.sel(time=tslice)
    if len(arr.time) == 0:
        raise DataCoverageError(f"No data available in period {period}")
    if season != "all":
        if not season in arr.season:
            raise DataCoverageError(f"No data available in {season} in period {period}")
        elif TsType(coldata.ts_type) < "monthly":
            raise TemporalResolutionError(
                "Season selection is only available for monthly or higher resolution data"
            )
        mask = arr["season"] == season
        arr = arr.sel(time=arr["time"][mask])

    return ColocatedData(arr)


def _process_heatmap_data(
    data,
    region_ids,
    use_weights,
    use_country,
    meta_glob,
    periods,
    seasons,
    add_trends,
    trends_min_yrs,
):
    output = {}
    stats_dummy = _init_stats_dummy()
    for freq, coldata in data.items():
        output[freq] = hm_freq = {}
        use_dummy = True if coldata is None else False
        for regid, regname in region_ids.items():
            hm_freq[regname] = {}
            for per in periods:
                for season in seasons:
                    perstr = f"{per}-{season}"
                    if use_dummy:
                        stats = stats_dummy
                    else:
                        try:
                            subset = _select_period_season_coldata(coldata, per, season)

                            trends_successful = False
                            if add_trends and freq != "daily":
                                # Calculates the start and stop years. min_yrs have a test value of 7 years. Should be set in cfg
                                (start, stop) = _get_min_max_year_periods([per])

                                if stop - start >= trends_min_yrs:
                                    try:
                                        subset_time_series = subset.get_regional_timeseries(
                                            regid, check_country_meta=use_country
                                        )

                                        (obs_trend, mod_trend) = _make_trends_from_timeseries(
                                            subset_time_series["obs"],
                                            subset_time_series["mod"],
                                            freq,
                                            season,
                                            start,
                                            stop,
                                            trends_min_yrs,
                                        )

                                        trends_successful = True
                                    except AeroValTrendsError as e:
                                        msg = f"Failed to calculate trends, and will skip. This was due to {e}"
                                        logger.warning(msg)

                            subset = subset.filter_region(
                                region_id=regid, check_country_meta=use_country
                            )

                            stats = _get_extended_stats(subset, use_weights)

                            if add_trends and freq != "daily" and trends_successful:
                                # The whole trends dicts are placed in the stats dict
                                stats["obs_trend"] = obs_trend
                                stats["mod_trend"] = mod_trend

                        except (DataCoverageError, TemporalResolutionError) as e:
                            stats = stats_dummy

                    hm_freq[regname][perstr] = stats

    return output


def _map_indices(outer_idx, inner_idx):
    """
    Find index positions of inner array contained in outer array

    Note
    ----
    Both outer and inner index arrays must be strictly monotonous

    Parameters
    ----------
    outer_idx : np.ndarray
        numerical value array (e.g. numerical timestamps in yearly resolution
        from 2010-2020). Must be strictly monotonous.
    inner_idx : np.ndarray
        inner value array, must be contained in outer array, e.g.
        numerical timestamps in yearly resolution from 2012-2014, all values
        in this array must be contained

    Returns
    -------
    mappingg : np.ndarray
        same shape as `outer_idx` with values -1 where no matches are found
        and else, corresponding indices of `inner_idx`

    Example
    -------
    `outer_idx`: [2010, 2011, 2012, 2013, 2014, 2015]
    `inner_idx`: [2011, 2012]
    `mapping`  : [-1, 0, 1, -1, -1, -1]

    """
    inner_start, inner_stop = inner_idx[0], inner_idx[-1]
    mapping = np.ones_like(outer_idx) * -1
    count = 0
    indata = False
    for i, idx in enumerate(outer_idx):
        if inner_start == idx:
            indata = True
        elif inner_stop < idx:
            break
        if indata:
            mapping[i] = count
            # sanity checking, will fail, e.g. if input data is not monotonuos
            assert inner_idx[count] == idx
            count += 1
    return mapping.astype(int)


def _process_statistics_timeseries(data, freq, region_ids, use_weights, use_country, data_freq):
    """
    Compute statistics timeseries for input data

    Parameters
    ----------
    data : dict
        dictionary containing colocated data object (values) in different
        temporal resolutions (keys).
    freq : str
        Output frequency (temporal resolution in which statistics timeseries
        if computed, AeroVal default is monthly)
    region_ids : dict
        Region IDs (keys) and corresponding names (values)
    use_weights : bool
        calculate statistics using area weights or not (only relevant for 4D
        colocated data with lat and lon dimension, e.g. from gridded / gridded
        co-location)
    use_country : bool
        Use countries for regional filtering.
    data_freq : str, optional
        Base frequency for computation of statistics (if None, `freq` is used).
        For details see https://github.com/metno/pyaerocom/pull/416.

    Raises
    ------
    TemporalResolutionError
        If `data_freq` is lower resolution than `freq`.

    Returns
    -------
    output : dict
        Dictionary with results.

    """
    if data_freq is None:
        data_freq = freq

    # input frequency is lower resolution than output frequency
    if TsType(data_freq) < TsType(freq):
        raise TemporalResolutionError(
            f"Desired input frequency {data_freq} is lower than desired "
            f"output frequency {freq}"
        )

    output = {}
    if not data_freq in data or data[data_freq] is None:
        raise TemporalResolutionError(
            f"failed to compute statistics timeseries, no co-located data "
            f"available in specified base resolution {data_freq}"
        )

    coldata = data[data_freq]

    # get time index of output frequency
    to_idx = data[freq].data.time.values
    tstr = TsType(freq).to_numpy_freq()
    # list of strings of output timestamps (used below to select the
    # individual periods)
    to_idx_str = [str(x) for x in to_idx.astype(f"datetime64[{tstr}]")]
    jsdate = _get_jsdate(to_idx)

    for regid, regname in region_ids.items():
        output[regname] = {}
        try:
            subset = coldata.filter_region(region_id=regid, check_country_meta=use_country)
        except DataCoverageError:
            continue
        for i, js in enumerate(jsdate):
            per = to_idx_str[i]
            try:
                arr = ColocatedData(subset.data.sel(time=per))
                stats = arr.calc_statistics(use_area_weights=use_weights)
                output[regname][str(js)] = _prep_stats_json(stats)
            except DataCoverageError:
                pass

    return output


def _get_jsdate(nparr):
    dt = nparr.astype("datetime64[s]")
    offs = np.datetime64("1970", "s")
    return (dt - offs).astype(int) * 1000


def _init_data_default_frequencies(coldata, to_ts_types):
    """
    Compute one colocated data object for each desired statistics frequency

    Parameters
    ----------
    coldata : ColocatedData
        Initial colocated data in a certain frequency
    to_ts_types : list
        list of desired frequencies for which statistical parameters are being
        computed.

    Returns
    -------
    dict
        keys are elements of `to_ts_types`, values are corresponding
        instances of `ColocatedData` in that resolution, or None, if resolution
        is higher than original resolution of input `coldata`.
    dict
        like `data_arrs` but values are jsdate instances.

    """

    data_arrs = dict.fromkeys(to_ts_types)

    from_tst = TsType(coldata.ts_type)

    for to in to_ts_types:
        to_tst = TsType(to)
        if from_tst < to_tst:
            continue
        elif from_tst == to_tst:
            cd = coldata.copy()
        else:
            cd = coldata.resample_time(to, settings_from_meta=True, inplace=False)
        # add season coordinate for later filtering
        arr = cd.data
        arr["season"] = arr.time.dt.season
        jsdates = _get_jsdate(arr.time.values)
        arr = arr.assign_coords(jsdate=("time", jsdates))
        cd.data = arr
        data_arrs[to] = cd

    return data_arrs


def _start_stop_from_periods(periods):
    start, stop = _get_min_max_year_periods(periods)
    return start_stop(start, stop + 1)



"""
General helper methods for the pyaerocom library.
"""
import logging
import re

import numpy as np

from pyaerocom.exceptions import TemporalResolutionError
from pyaerocom.time_config import (
    PANDAS_FREQ_TO_TS_TYPE,
    TS_TYPE_TO_NUMPY_FREQ,
    TS_TYPE_TO_PANDAS_FREQ,
    TS_TYPE_TO_SI,
    TS_TYPES,
)

logger = logging.getLogger(__name__)


class TsType:
    VALID = TS_TYPES
    VALID_ITER = VALID[:-1]
    FROM_PANDAS = PANDAS_FREQ_TO_TS_TYPE
    TO_PANDAS = TS_TYPE_TO_PANDAS_FREQ
    TO_NUMPY = TS_TYPE_TO_NUMPY_FREQ
    TO_SI = TS_TYPE_TO_SI

    TS_MAX_VALS = {
        "minutely": 360,  # up to 6hourly
        "hourly": 168,  # up to weekly
        "daily": 180,  # up to 6 monthly
        "weekly": 104,  # up to ~2yearly
        "monthly": 120,
    }  # up to 10yearly

    TSTR_TO_CF = {"hourly": "hours", "daily": "days", "monthly": "days"}

    TOL_SECS_PERCENT = 5

    def __init__(self, val):
        self._mulfac = 1
        self._val = None

        self.val = val

    @property
    def mulfac(self):
        """Multiplication factor of frequency"""
        return self._mulfac

    @mulfac.setter
    def mulfac(self, value):
        try:
            value = int(value)
        except Exception:
            raise ValueError("mulfac needs to be int or convertible to int")
        if self.base in self.TS_MAX_VALS and value > self.TS_MAX_VALS[self.base]:
            raise ValueError(
                f"Multiplication factor exceeds maximum allowed, which is "
                f"{self.TS_MAX_VALS[self.base]}"
            )
        self._mulfac = value

    @property
    def base(self):
        """Base string (without multiplication factor, cf :attr:`mulfac`)"""
        return self._val

    @property
    def val(self):
        """Value of frequency (string type), e.g. 3daily"""
        if self._mulfac != 1:
            return f"{self._mulfac}{self._val}"
        return self._val

    @val.setter
    def val(self, val):
        if val is None:
            raise TemporalResolutionError(
                "Invalid input, please provide valid frequency string..."
            )
        mulfac = 1
        if val[0].isdigit():
            ivalstr = re.findall(r"\d+", val)[0]
            val = val.split(ivalstr)[-1]
            mulfac = int(ivalstr)
        if not val in self.VALID:
            try:
                val = self._from_pandas(val)
            except TemporalResolutionError:
                raise TemporalResolutionError(
                    f"Invalid input for ts_type {val}. Choose from {self.VALID}"
                )
        if val in self.TS_MAX_VALS and mulfac != 1:
            if mulfac > self.TS_MAX_VALS[val]:
                raise TemporalResolutionError(
                    f"Invalid input for ts_type {val}. Multiplication factor "
                    f"{mulfac} exceeds maximum allowed for {val}, which is "
                    f"{self.TS_MAX_VALS[val]}"
                )
        self._val = val
        self._mulfac = mulfac

    @property
    def datetime64_str(self):
        """Convert ts_type str to datetime64 unit string"""
        return f"datetime64[{self.to_numpy_freq()}]"

    @property
    def timedelta64_str(self):
        """Convert ts_type str to datetime64 unit string"""
        return f"timedelta64[{self.to_numpy_freq()}]"

    @property
    def cf_base_unit(self):
        """Convert ts_type str to CF convention time unit"""
        if not self.base in self.TSTR_TO_CF:
            raise NotImplementedError(f"Cannot convert {self.base} to CF str")
        return self.TSTR_TO_CF[self.base]

    @property
    def num_secs(self):
        """Number of seconds in one period

        Note
        ----
        Be aware that for monthly frequency the number of seconds is not well
        defined!
        """
        from cf_units import Unit

        cf = self.to_si()
        total_secs = 1 / Unit("s").convert(1, cf)
        return total_secs

    @property
    def tol_secs(self):
        """Tolerance in seconds for current TsType"""
        total_secs = self.num_secs
        frac = self.TOL_SECS_PERCENT / 100
        return int(np.ceil(frac * total_secs))

    def to_timedelta64(self):
        """
        Convert frequency to timedelta64 object

        Can be used, e.g. as tolerance when reindexing pandas Series

        Returns
        -------
        timedelta64

        """
        return np.timedelta64(1, self.to_numpy_freq())

    @property
    def next_higher(self):
        """Next lower resolution code"""
        if self.mulfac > 1:
            return TsType(self._val)

        idx = self.VALID_ITER.index(self._val)
        if idx == 0:
            raise IndexError(f"No higher resolution available than {self}")
        return TsType(self.VALID_ITER[idx - 1])

    @property
    def next_lower(self):
        """Next lower resolution code

        This will go to the next lower base resolution, that is if current is
        3daily, it will return weekly, however, if current exceeds next lower
        base, it will iterate that base, that is, if current is 8daily, next
        lower will be 2weekly (and not 9daily).
        """
        idx = self.VALID_ITER.index(self._val)
        if idx == len(self.VALID_ITER) - 1:
            tst = TsType(self.base)
            tst.mulfac = self.mulfac + 1
            return tst
        tst = TsType(self.VALID_ITER[idx + 1])
        if self.mulfac == 1 or self.num_secs < tst.num_secs:
            return tst
        try:
            maxmul = self.TS_MAX_VALS[tst.base]
        except:
            maxmul = 10
        numsecs = self.num_secs
        for mulfac in range(1, maxmul + 1):
            tst.mulfac = mulfac
            if numsecs < tst.num_secs:
                return tst
        raise TemporalResolutionError(f"Failed to determine next lower resolution for {self}")

    @staticmethod
    def valid(val):
        try:
            TsType(val)
            return True
        except TemporalResolutionError:
            return False

    def to_numpy_freq(self):
        if not self._val in self.TO_NUMPY:
            raise TemporalResolutionError(f"numpy frequency not available for {self._val}")
        freq = self.TO_NUMPY[self._val]
        return f"{self.mulfac}{freq}"

    def to_pandas_freq(self):
        """Convert ts_type to pandas frequency string"""
        if not self._val in self.TO_PANDAS:
            raise TemporalResolutionError(f"pandas frequency not available for {self._val}")
        freq = self.TO_PANDAS[self._val]
        if self._mulfac == 1:
            return freq
        return f"{self._mulfac}{freq}"

    def to_si(self):
        """Convert to SI conform string (e.g. used for unit conversion)"""
        base = self.base
        if not base in self.TO_SI:
            raise ValueError(f"Cannot convert ts_type={self} to SI unit string...")
        si = self.TO_SI[base]
        return si if self.mulfac == 1 else f"({self.mulfac}{si})"

    def get_min_num_obs(self, to_ts_type: "TsType", min_num_obs: dict) -> int:
        selfstr = self.val
        if to_ts_type >= self:  # should occur rarely
            if to_ts_type == self:
                return 0
            raise TemporalResolutionError(
                f"input ts_type {to_ts_type} is lower resolution than current {self}"
            )

        elif str(to_ts_type) in min_num_obs:
            # output frequency is specified in min_num_obs (note: this may
            # also be 3daily, etc, i.e., not restricted to base frequencies)
            mno = min_num_obs[str(to_ts_type)]
            if selfstr in mno:
                return int(mno[selfstr])
            elif self.mulfac != 1 and self.base in mno:
                min_num_base = mno[self.base]
                return int(np.round(min_num_base / self.mulfac))

        elif to_ts_type.base in min_num_obs:
            mno = min_num_obs[to_ts_type.base]
            if selfstr in mno:
                val = mno[selfstr]
                return int(np.round(to_ts_type.mulfac * val))

            elif self.mulfac != 1 and self.base in mno:
                min_num_base = mno[self.base]
                val = min_num_base / self.mulfac * to_ts_type.mulfac
                val = int(np.round(val))
                return val
        raise ValueError(
            f"could not infer min_num_obs value from input dict {min_num_obs} "
            f"for conversion from {self} to {to_ts_type}"
        )

    def check_match_total_seconds(self, total_seconds):
        """
        Check if this object matches with input interval length in seconds

        Parameters
        ----------
        total_seconds : int or float
            interval length in units of seconds (e.g. 86400 for daily)

        Returns
        -------
        bool

        """
        try:
            numsecs = self.num_secs
            tolsecs = self.tol_secs
        except ValueError:  # native / undefined
            return False
        low, high = numsecs - tolsecs, numsecs + tolsecs
        if np.logical_and(total_seconds >= low, total_seconds <= high):
            return True
        return False

    @staticmethod
    def _try_infer_from_total_seconds(base, total_seconds):
        """
        Infer multiplication factor required to match input interval length

        Not to be used directly, is used in :func:`from_total_seconds`.

        Parameters
        ----------
        base : str
            base frequency
        total_seconds : int or float
            interval length

        Raises
        ------
        TemporalResolutionError
            if TsType cannot be inferred

        Returns
        -------
        TsType
            inferred frequency

        """

        if base in TsType.TS_MAX_VALS:
            maxnum = TsType.TS_MAX_VALS[base]
        else:
            maxnum = 2
        candidates = []
        dts = []
        tstype = TsType(base)
        for mulfac in range(1, maxnum):
            tstype.mulfac = mulfac
            if tstype.check_match_total_seconds(total_seconds):
                dt = total_seconds - tstype.num_secs
                dts.append(dt)
                candidates.append(TsType(f"{mulfac}{base}"))
                if dt == 0 or dt < 0:  # current candidate has larger number of seconds than input
                    break

        if len(candidates) > 0:
            return candidates[np.argmin(np.abs(dts))]

        raise TemporalResolutionError(
            f"Period {total_seconds}s could not be associated with any "
            f"allowed multiplication factor of base frequency {base}"
        )

    @staticmethod
    def from_total_seconds(total_seconds):
        """
        Try to infer TsType based on interval length

        Parameters
        ----------
        total_seconds : int or float
            total number of seconds

        Raises
        ------
        TemporalResolutionError
            If no TsType can be inferred for input number of seconds

        Returns
        -------
        TsType

        """
        candidates = []
        candidates_diff = []
        for tst in TsType.VALID_ITER:
            tstype = TsType(tst)
            if tstype.check_match_total_seconds(total_seconds):
                return tstype
            diff = total_seconds - tstype.num_secs
            if diff > 0:
                candidates.append(tst)
                candidates_diff.append(diff)
        if len(candidates) > 0:
            # sort by the candidate that has the lowest dt
            candidates_sorted = [c for _, c in sorted(zip(candidates_diff, candidates))]
            for base_tst in candidates_sorted:
                try:
                    return TsType._try_infer_from_total_seconds(base_tst, total_seconds)
                except TemporalResolutionError as e:
                    logger.info(e)
                    continue

        raise TemporalResolutionError(
            f"failed to infer ts_type based on input dt={total_seconds} s"
        )

    def _from_pandas(self, val):
        if not val in self.FROM_PANDAS:
            raise TemporalResolutionError(f"Invalid input: {val}, need pandas frequency string")
        return self.FROM_PANDAS[val]

    def __eq__(self, other):
        if isinstance(other, str):
            other = TsType(other)
        return other.val == self.val

    def __lt__(self, other):
        if isinstance(other, str):
            other = TsType(other)
        nss, nso = self.num_secs, other.num_secs
        # inverted comparison, i.e. if other has less seconds if has higher
        # resolution
        return nss > nso

    def __le__(self, other):
        return True if (self.__eq__(other) or self.__lt__(other)) else False

    def __gt__(self, other):
        return not self.__le__(other)

    def __ge__(self, other):
        return not self.__lt__(other)

    def __call__(self):
        return self.val

    def __str__(self):
        return self.val

    def __repr__(self):
        return self.val
